"""Device and module mappings for VeSync devices.

**To add a new device type to existing module: Add the device_type to the end
of the existing dev_types list.**

This module contains mappings for VeSync devices to their respective classes.
The mappings are used to create instances of the appropriate device class
based on the device type and define features and modes. The device type is taken
from the `deviceType` field in the device list API.

The AirFryerMap, OutletMap, SwitchMap, BulbMap, FanMap, HumidifierMap, PurifierMap
and ThermostatMap dataclasses are used to define the mappings for each product type
with the associated module, class, features and other device specific configuration. The
[`get_device_config`][pyvesync.device_map.get_device_config] function is used
to get the device map object from the device type to instantiate the appropriate class.
The individual `get_<product-type>` functions are used to get the device details
for the specific to the device type. Both functions return the same model,
the individual product type functions are used to satisfy type checking in
the individual devices.

Attributes:
    outlet_modules: list[OutletMap]: List of Outlet device mappings.
    switch_modules: list[SwitchMap]: List of Switch device mappings.
    bulb_modules: list[BulbMap]: List of Bulb device mappings.
    fan_modules: list[FanMap]: List of Fan device mappings.
    purifier_modules: list[PurifierMap]: List of Purifier device mappings.
    humidifier_modules: list[HumidifierMap]: List of Humidifier device mappings.
    air_fryer_modules: list[AirFryerMap]: List of Air Fryer device mappings.
    thermostat_modules: list[ThermostatMap]: List of Thermostat device mappings.

Classes:
    ProductTypes: Enum: General device types enum.
    DeviceMapTemplate: Template for DeviceModules mapping.
    OutletMap: Template for Outlet device mapping.
    SwitchMap: Template for Switch device mapping.
    BulbMap: dataclass: Template for Bulb device mapping.
    FanMap: dataclass: Template for Fan device mapping.
    HumidifierMap: dataclass: Template for Humidifier device mapping.
    PurifierMap: dataclass: Template for Purifier device mapping.
    AirFryerMap: dataclass: Template for Air Fryer device mapping.
    ThermostatMap: dataclass: Template for Thermostat device mapping.

Functions:
    get_device_config: Get the device map object from the device type.
    get_outlet: Get outlet device config, returning OutletMap object.
    get_switch: Get switch device config, returning SwitchMap object.
    get_bulb: Get the bulb device config, returning BulbMap object.
    get_fan: Get the fan device config, returning the FanMap object.
    get_humidifier: Get the humidfier config, returning the HumidifierMap object.
    get_purifier: Get the purifier config, returning the PurifierMap object.
    get_air_fryer: Get the Air Fryer config, returning the AirFryerMap object.
    get_thermostat: Get the thermostat config, returning the ThermostatMap object.

Note:
    To add devices, add the device mapping to the appropriate `<product-type>_modules`
    list, ensuring all required fields are present based on the `<product-type>Map`
    fields.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from itertools import chain
from types import ModuleType
from typing import Union

from pyvesync.const import (
    BulbFeatures,
    ColorMode,
    EnergyIntervals,
    FanFeatures,
    FanModes,
    FanSleepPreference,
    HumidifierFeatures,
    HumidifierModes,
    NightlightModes,
    OutletFeatures,
    ProductLines,
    ProductTypes,
    PurifierAutoPreference,
    PurifierFeatures,
    PurifierModes,
    SwitchFeatures,
    ThermostatEcoTypes,
    ThermostatFanModes,
    ThermostatHoldOptions,
    ThermostatRoutineTypes,
    ThermostatWorkModes,
)
from pyvesync.devices import (
    vesyncbulb,
    vesyncfan,
    vesynchumidifier,
    vesynckitchen,
    vesyncoutlet,
    vesyncpurifier,
    vesyncswitch,
    vesyncthermostat,
)

T_MAPS = Union[  # noqa: UP007, RUF100
    list['OutletMap'],
    list['SwitchMap'],
    list['BulbMap'],
    list['FanMap'],
    list['HumidifierMap'],
    list['PurifierMap'],
    list['AirFryerMap'],
    list['ThermostatMap'],
]


@dataclass(kw_only=True)
class DeviceMapTemplate:
    """Template for DeviceModules mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device.
        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
    """

    dev_types: list[str]
    class_name: str
    product_type: str
    product_line: str
    module: ModuleType
    setup_entry: str
    model_display: str
    model_name: str
    device_alias: str | None = None
    features: list[str] = field(default_factory=list)


@dataclass(kw_only=True)
class OutletMap(DeviceMapTemplate):
    """Template for DeviceModules mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device - ProductTypes.OUTLET
        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
        product_type (str): Product type of the device.
        module (ModuleType): Module for the device.
        nightlight_modes (list[str]): List of nightlight modes for the device.
    """

    product_line: str = ProductLines.WIFI_LIGHT
    product_type: str = ProductTypes.OUTLET
    module: ModuleType = vesyncoutlet
    energy_intervals: tuple[str, ...] = (
        EnergyIntervals.YEAR,
        EnergyIntervals.MONTH,
        EnergyIntervals.WEEK,
    )
    nightlight_modes: list[NightlightModes] = field(default_factory=list)


@dataclass(kw_only=True)
class SwitchMap(DeviceMapTemplate):
    """Template for DeviceModules mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device - ProductTypes.SWITCH
        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
        product_type (str): Product type of the device.
        module (ModuleType): Module for the device.
    """

    product_line: str = ProductLines.SWITCHES
    product_type: str = ProductTypes.SWITCH
    module: ModuleType = vesyncswitch


@dataclass(kw_only=True)
class BulbMap(DeviceMapTemplate):
    """Template for DeviceModules mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device - ProductTypes.BULB
        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
        color_model (str | None): Color model for the device.
        color_modes (list[str]): List of color modes for the device.
    """

    product_line: str = ProductLines.WIFI_LIGHT
    color_model: str | None = None
    product_type: str = ProductTypes.BULB
    module: ModuleType = vesyncbulb
    color_modes: list[str] = field(default_factory=list)


@dataclass(kw_only=True)
class FanMap(DeviceMapTemplate):
    """Template for DeviceModules mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device - ProductTypes.FAN
        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
        fan_levels (list[int]): List of fan levels for the device.
        modes (list[str]): List of modes for the device.
        sleep_preferences (list[str]): List of sleep preferences for the device.
        set_mode_method (str): Method to set the mode for the device.
    """

    product_line: str = ProductLines.WIFI_AIR
    product_type: str = ProductTypes.FAN
    module: ModuleType = vesyncfan
    fan_levels: list[int] = field(default_factory=list)
    modes: dict[str, str] = field(default_factory=dict)
    sleep_preferences: list[str] = field(default_factory=list)
    set_mode_method: str = ''


@dataclass(kw_only=True)
class HumidifierMap(DeviceMapTemplate):
    """Template for DeviceModules mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device - ProductTypes.HUMIDIFIER
        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
        mist_modes (dict[str, str]): Dictionary of mist modes for the device.
        mist_levels (list[int | str]): List of mist levels for the device.
        target_minmax (tuple[int, int]): Minimum and maximum target humidity levels.
        warm_mist_levels (list[int | str]): List of warm mist levels for the device.
    """

    product_line: str = ProductLines.WIFI_AIR
    mist_modes: dict[str, str] = field(default_factory=dict)
    mist_levels: list[int] = field(default_factory=list)
    product_type: str = ProductTypes.HUMIDIFIER
    module: ModuleType = vesynchumidifier
    target_minmax: tuple[int, int] = (30, 80)
    warm_mist_levels: list[int] = field(default_factory=list)


@dataclass(kw_only=True)
class PurifierMap(DeviceMapTemplate):
    """Template for DeviceModules mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device - ProductTypes.PURIFIER

        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
        fan_levels (list[int]): List of fan levels for the device.
        modes (list[str]): List of modes for the device.
        nightlight_modes (list[str]): List of nightlight modes for the device.
        auto_preferences (list[str]): List of auto preferences for the device.
    """

    product_line: str = ProductLines.WIFI_AIR
    product_type: str = ProductTypes.PURIFIER
    module: ModuleType = vesyncpurifier
    fan_levels: list[int] = field(default_factory=list)
    modes: list[str] = field(default_factory=list)
    nightlight_modes: list[str] = field(default_factory=list)
    auto_preferences: list[str] = field(default_factory=list)


@dataclass(kw_only=True)
class AirFryerMap(DeviceMapTemplate):
    """Template for DeviceModules mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device - ProductTypes.AIR_FRYER
        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
        product_type (str): Product type of the device.
        module (ModuleType): Module for the device.
    """

    temperature_range_f: tuple[int, int] = (200, 400)
    temperature_range_c: tuple[int, int] = (75, 200)
    product_line: str = ProductLines.WIFI_KITCHEN
    product_type: str = ProductTypes.AIR_FRYER
    module: ModuleType = vesynckitchen


@dataclass(kw_only=True)
class ThermostatMap(DeviceMapTemplate):
    """Template for Thermostat device mapping.

    Attributes:
        dev_types (list[str]): List of device types to match from API.
        class_name (str): Class name of the device.
        product_type (str): Product type of the device.
        module (ModuleType): Module for the device.
        setup_entry (str): Setup entry for the device, if unknown use the device_type
            base without region
        model_display (str): Display name of the model.
        model_name (str): Name of the model.
        device_alias (str | None): Alias for the device, if any.
        features (list[str]): List of features for the device.
        modes (list[int]): List of modes for the device.
        fan_modes (list[int]): List of fan modes for the device.
        eco_types (list[int]): List of eco types for the device.
        hold_options (list[int]): List of hold options for the device.
        routine_types (list[int]): List of routine types for the device.
    """

    product_line: str = ProductLines.THERMOSTAT
    product_type: str = ProductTypes.THERMOSTAT
    module: ModuleType = vesyncthermostat
    modes: list[int] = field(default_factory=list)
    fan_modes: list[int] = field(default_factory=list)
    eco_types: list[int] = field(default_factory=list)
    hold_options: list[int] = field(default_factory=list)
    routine_types: list[int] = field(default_factory=list)


thermostat_modules = [
    ThermostatMap(
        dev_types=['LTM-A401S-WUS'],
        class_name='VeSyncAuraThermostat',
        fan_modes=[
            ThermostatFanModes.AUTO,
            ThermostatFanModes.CIRCULATE,
            ThermostatFanModes.ON,
        ],
        modes=[
            ThermostatWorkModes.HEAT,
            ThermostatWorkModes.COOL,
            ThermostatWorkModes.AUTO,
            ThermostatWorkModes.OFF,
            ThermostatWorkModes.SMART_AUTO,
            ThermostatWorkModes.EM_HEAT,
        ],
        eco_types=[
            ThermostatEcoTypes.BALANCE,
            ThermostatEcoTypes.COMFORT_FIRST,
            ThermostatEcoTypes.COMFORT_SECOND,
            ThermostatEcoTypes.ECO_FIRST,
            ThermostatEcoTypes.ECO_SECOND,
        ],
        hold_options=[
            ThermostatHoldOptions.PERMANENTLY,
            ThermostatHoldOptions.FOUR_HOURS,
            ThermostatHoldOptions.TWO_HOURS,
            ThermostatHoldOptions.UNTIL_NEXT_SCHEDULED_ITEM,
        ],
        routine_types=[
            ThermostatRoutineTypes.AWAY,
            ThermostatRoutineTypes.CUSTOM,
            ThermostatRoutineTypes.HOME,
            ThermostatRoutineTypes.SLEEP,
        ],
        setup_entry='LTM-A401S-WUS',
        model_display='LTM-A401S Series',
        model_name='Aura Thermostat',
    )
]


outlet_modules = [
    OutletMap(
        dev_types=['wifi-switch-1.3'],
        class_name='VeSyncOutlet7A',
        features=[OutletFeatures.ENERGY_MONITOR],
        model_name='WiFi Outlet US/CA',
        model_display='ESW01-USA Series',
        setup_entry='wifi-switch-1.3',
    ),
    OutletMap(  # TODO: Add energy
        dev_types=['ESW10-USA', 'ESW10-EU'],
        class_name='VeSyncESW10USA',
        features=[],
        model_name='10A WiFi Outlet USA',
        model_display='ESW10-USA Series',
        setup_entry='ESW10-USA',
    ),
    OutletMap(
        dev_types=['ESW01-EU', 'ESW01-USA', 'ESW03-USA', 'ESW03-EU'],
        class_name='VeSyncOutlet10A',
        features=[OutletFeatures.ENERGY_MONITOR],
        model_name='ESW03 10A WiFi Outlet',
        model_display='ESW01/03 USA/EU',
        setup_entry='ESW03',
    ),
    OutletMap(
        dev_types=['ESW15-USA'],
        class_name='VeSyncOutlet15A',
        features=[OutletFeatures.ENERGY_MONITOR, OutletFeatures.NIGHTLIGHT],
        nightlight_modes=[NightlightModes.ON, NightlightModes.OFF, NightlightModes.AUTO],
        model_name='15A WiFi Outlet US/CA',
        model_display='ESW15-USA Series',
        setup_entry='ESW15-USA',
    ),
    OutletMap(
        dev_types=['ESO15-TB'],
        class_name='VeSyncOutdoorPlug',
        features=[OutletFeatures.ENERGY_MONITOR],
        model_name='Outdoor Plug',
        model_display='ESO15-TB Series',
        setup_entry='ESO15-TB',
    ),
    OutletMap(
        dev_types=[
            'WHOGPLUG',
        ],
        class_name='VeSyncOutletWHOGPlug',
        features=[OutletFeatures.ONOFF, OutletFeatures.ENERGY_MONITOR],
        model_name='Smart Plug',
        model_display='Smart Plug Series',
        setup_entry='WHOGPLUG',
        device_alias='Greensun Smart Plug',
    ),
    OutletMap(
        dev_types=[
            'BSDOG01',
            'BSDOG02',
            'WYSMTOD16A',
            'WM-PLUG',
            'JXUK13APLUG',
            'WYZYOGMINIPLUG',
            'HWPLUG16A',
            'FY-PLUG',
            'HWPLUG16',
        ],
        class_name='VeSyncBSDOGPlug',
        features=[OutletFeatures.ONOFF, OutletFeatures.ENERGY_MONITOR],
        model_name='Smart Plug',
        model_display='Smart Plug Series',
        setup_entry='BSDOG01',
        device_alias='Smart Plug Series',
    ),
]
"""List of ['OutletMap'][pyvesync.device_map.OutletMap] configuration
objects for outlet devices."""

switch_modules = [
    SwitchMap(
        dev_types=['ESWL01'],
        class_name='VeSyncWallSwitch',
        device_alias='Wall Switch',
        features=[SwitchFeatures.ONOFF],
        model_name='Light Switch',
        model_display='ESWL01 Series',
        setup_entry='ESWL01',
    ),
    SwitchMap(
        dev_types=['ESWD16'],
        class_name='VeSyncDimmerSwitch',
        features=[
            SwitchFeatures.DIMMABLE,
            SwitchFeatures.INDICATOR_LIGHT,
            SwitchFeatures.BACKLIGHT_RGB,
        ],
        device_alias='Dimmer Switch',
        model_name='Dimmer Switch',
        model_display='ESWD16 Series',
        setup_entry='ESWD16',
    ),
    SwitchMap(
        dev_types=['ESWL03'],
        class_name='VeSyncWallSwitch',
        device_alias='Three-Way Wall Switch',
        features=[SwitchFeatures.ONOFF],
        model_name='Light Switch 3 way',
        model_display='ESWL03 Series',
        setup_entry='ESWL03',
    ),
]
"""List of ['SwitchMap'][pyvesync.device_map.SwitchMap] configuration
objects for switch devices."""


bulb_modules = [
    BulbMap(
        dev_types=['ESL100'],
        class_name='VeSyncBulbESL100',
        features=[SwitchFeatures.DIMMABLE],
        color_model=None,
        device_alias='Dimmable Bright White Bulb',
        color_modes=[ColorMode.WHITE],
        model_display='ESL100 Series',
        model_name='Soft white light bulb',
        setup_entry='ESL100',
    ),
    BulbMap(
        dev_types=['ESL100CW'],
        class_name='VeSyncBulbESL100CW',
        features=[BulbFeatures.DIMMABLE, BulbFeatures.COLOR_TEMP],
        color_model=None,
        device_alias='Dimmable Tunable White Bulb',
        color_modes=[ColorMode.WHITE],
        model_display='ESL100CW Series',
        model_name='Cool-to-Warm White Light Bulb',
        setup_entry='ESL100CW',
    ),
    BulbMap(
        dev_types=['XYD0001'],
        class_name='VeSyncBulbValcenoA19MC',
        features=[
            BulbFeatures.DIMMABLE,
            BulbFeatures.MULTICOLOR,
            BulbFeatures.COLOR_TEMP,
        ],
        color_model=ColorMode.HSV,
        device_alias='Valceno Dimmable RGB Bulb',
        color_modes=[ColorMode.WHITE, ColorMode.COLOR],
        model_display='XYD0001',
        model_name='Valceno WiFi Bulb',
        setup_entry='XYD0001',
    ),
    BulbMap(
        dev_types=['ESL100MC'],
        class_name='VeSyncBulbESL100MC',
        features=[
            BulbFeatures.MULTICOLOR,
            BulbFeatures.DIMMABLE,
        ],
        color_model=ColorMode.RGB,
        device_alias='Etekcity Dimmable RGB Bulb',
        color_modes=[ColorMode.WHITE, ColorMode.COLOR],
        model_name='Multicolor Bulb',
        model_display='ESL100MC',
        setup_entry='ESL100MC',
    ),
]
"""List of ['BulbMap'][pyvesync.device_map.BulbMap] configuration
objects for bulb devices."""


humidifier_modules = [
    HumidifierMap(
        class_name='VeSyncHumid200300S',
        dev_types=[
            'Classic300S',
            'LUH-A601S-WUSB',
            'LUH-A601S-AUSW',
        ],
        features=[
            HumidifierFeatures.NIGHTLIGHT,
            HumidifierFeatures.NIGHTLIGHT_BRIGHTNESS,
            HumidifierFeatures.AUTO_STOP,
        ],
        mist_modes={
            HumidifierModes.AUTO: 'auto',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 10)),
        device_alias='Classic 300S',
        model_display='LUH-A601S Series',
        model_name='Classic 300S',
        setup_entry='Classic300S',
    ),
    HumidifierMap(
        class_name='VeSyncHumid200S',
        dev_types=['Classic200S'],
        features=[HumidifierFeatures.AUTO_STOP],
        mist_modes={
            HumidifierModes.AUTO: 'auto',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 10)),
        device_alias='Classic 200S',
        model_display='Classic 200S Series',
        model_name='Classic 200S',
        setup_entry='Classic200S',
    ),
    HumidifierMap(
        class_name='VeSyncHumid200300S',
        dev_types=[
            'Dual200S',
            'LUH-D301S-WUSR',
            'LUH-D301S-WJP',
            'LUH-D301S-WEU',
            'LUH-D301S-KEUR',
        ],
        features=[HumidifierFeatures.AUTO_STOP],
        mist_modes={
            HumidifierModes.AUTO: 'auto',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 3)),
        device_alias='Dual 200S',
        model_display='LUH-D301S Series',
        model_name='Dual 200S',
        setup_entry='Dual200S',
    ),
    HumidifierMap(
        class_name='VeSyncHumid200300S',
        dev_types=[
            'LUH-A602S-WUSR',
            'LUH-A602S-WUS',
            'LUH-A602S-WEUR',
            'LUH-A602S-WEU',
            'LUH-A602S-WJP',
            'LUH-A602S-WUSC',
        ],
        features=[HumidifierFeatures.WARM_MIST, HumidifierFeatures.AUTO_STOP],
        mist_modes={
            HumidifierModes.AUTO: 'auto',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 10)),
        warm_mist_levels=[0, 1, 2, 3],
        device_alias='LV600S',
        model_display='LUH-A602S Series',
        model_name='LV600S',
        setup_entry='LUH-A602S-WUS',
    ),
    HumidifierMap(
        class_name='VeSyncLV600S',
        dev_types=[
            'LUH-A603S-WUS',
        ],
        features=[HumidifierFeatures.WARM_MIST],
        mist_modes={
            HumidifierModes.AUTO: 'humidity',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 10)),
        warm_mist_levels=[0, 1, 2, 3],
        device_alias='LV600S',
        model_display='LUH-A603S Series',
        model_name='LV600S',
        setup_entry='LUH-A603S-WUS',
    ),
    HumidifierMap(
        class_name='VeSyncHumid200300S',
        dev_types=['LUH-O451S-WEU'],
        features=[HumidifierFeatures.WARM_MIST, HumidifierFeatures.AUTO_STOP],
        mist_modes={
            HumidifierModes.AUTO: 'auto',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 10)),
        warm_mist_levels=list(range(4)),
        device_alias='OasisMist 450S EU',
        model_name='OasisMist 4.5L',
        model_display='LUH-O451S Series',
        setup_entry='LUH-O451S-WEU',
    ),
    HumidifierMap(
        class_name='VeSyncHumid200300S',
        dev_types=['LUH-O451S-WUS', 'LUH-O451S-WUSR', 'LUH-O601S-WUS', 'LUH-O601S-KUS'],
        features=[HumidifierFeatures.WARM_MIST, HumidifierFeatures.AUTO_STOP],
        mist_modes={
            HumidifierModes.AUTO: 'auto',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.MANUAL: 'manual',
            HumidifierModes.HUMIDITY: 'humidity',
        },
        mist_levels=list(range(1, 10)),
        warm_mist_levels=list(range(4)),
        device_alias='OasisMist 450S',
        model_display='OasisMist 4.5L Series',
        model_name='OasisMist 4.5L',
        setup_entry='LUH-O451S-WUS',
    ),
    HumidifierMap(
        class_name='VeSyncHumid1000S',
        dev_types=['LUH-M101S-WUS', 'LUH-M101S-WUSR'],
        features=[HumidifierFeatures.AUTO_STOP],
        mist_modes={
            HumidifierModes.AUTO: 'auto',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 10)),
        device_alias='Oasismist 1000S',
        model_display='Oasismist Series',
        model_name='Oasismist 1000S',
        setup_entry='LUH-M101S-WUS',
    ),
    HumidifierMap(
        class_name='VeSyncHumid1000S',
        dev_types=['LUH-M101S-WEUR'],
        features=[
            HumidifierFeatures.NIGHTLIGHT,
            HumidifierFeatures.NIGHTLIGHT_BRIGHTNESS,
            HumidifierFeatures.AUTO_STOP,
        ],
        mist_modes={
            HumidifierModes.AUTO: 'auto',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 10)),
        device_alias='Oasismist 1000S EU',
        model_display='Oasismist Series EU',
        model_name='Oasismist 1000S EU',
        setup_entry='LUH-M101S-WEUR',
    ),
    HumidifierMap(
        class_name='VeSyncSuperior6000S',
        dev_types=['LEH-S601S-WUS', 'LEH-S601S-WUSR', 'LEH-S601S-WEUR', 'LEH-S602S-WUS'],
        features=[HumidifierFeatures.DRYING_MODE, HumidifierFeatures.AUTO_STOP],
        mist_modes={
            HumidifierModes.AUTO: 'autoPro',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.HUMIDITY: 'humidity',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 10)),
        device_alias='Superior 6000S',
        model_display='LEH-S601S Series',
        model_name='Superior 6000S',
        setup_entry='LEH-S601S',
    ),
    HumidifierMap(
        class_name='VeSyncSproutHumid',
        dev_types=['LEH-B381S-WUS', 'LEH-B381S-WEU'],
        features=[
            HumidifierFeatures.DRYING_MODE,
            HumidifierFeatures.NIGHTLIGHT,
            HumidifierFeatures.NIGHTLIGHT_BRIGHTNESS,
            HumidifierFeatures.AUTO_STOP,
        ],
        mist_modes={
            HumidifierModes.AUTO: 'autoPro',
            HumidifierModes.SLEEP: 'sleep',
            HumidifierModes.MANUAL: 'manual',
        },
        mist_levels=list(range(1, 3)),
        device_alias='Sprout Humidifier',
        model_display='LEH-B381S Series',
        model_name='Sprout Humidifier',
        setup_entry='LEH-B381S',
    ),
]
"""List of ['HumidifierMap'][pyvesync.device_map.HumidifierMap] configuration
objects for humidifier devices."""


purifier_modules: list[PurifierMap] = [
    PurifierMap(
        class_name='VeSyncAirBypass',
        dev_types=['Core200S', 'LAP-C201S-AUSR', 'LAP-C202S-WUSR'],
        modes=[PurifierModes.SLEEP, PurifierModes.MANUAL],
        features=[PurifierFeatures.RESET_FILTER, PurifierFeatures.NIGHTLIGHT],
        auto_preferences=[
            PurifierAutoPreference.DEFAULT,
            PurifierAutoPreference.EFFICIENT,
            PurifierAutoPreference.QUIET,
        ],
        fan_levels=list(range(1, 4)),
        nightlight_modes=[NightlightModes.ON, NightlightModes.OFF, NightlightModes.DIM],
        device_alias='Core 200S',
        model_display='Core 200S',
        model_name='Core 200S',
        setup_entry='Core200S',
    ),
    PurifierMap(
        class_name='VeSyncAirBypass',
        dev_types=[
            'Core300S',
            'LAP-C301S-WJP',
            'LAP-C302S-WUSB',
            'LAP-C301S-WAAA',
            'LAP-C302S-WGC',
        ],
        modes=[PurifierModes.SLEEP, PurifierModes.MANUAL, PurifierModes.AUTO],
        auto_preferences=[
            PurifierAutoPreference.DEFAULT,
            PurifierAutoPreference.EFFICIENT,
            PurifierAutoPreference.QUIET,
        ],
        features=[PurifierFeatures.AIR_QUALITY],
        fan_levels=list(range(1, 4)),
        device_alias='Core 300S',
        model_display='Core 300S',
        model_name='Core 300S',
        setup_entry='Core300S',
    ),
    PurifierMap(
        class_name='VeSyncAirBypass',
        dev_types=['Core400S', 'LAP-C401S-WJP', 'LAP-C401S-WUSR', 'LAP-C401S-WAAA'],
        modes=[PurifierModes.SLEEP, PurifierModes.MANUAL, PurifierModes.AUTO],
        features=[PurifierFeatures.AIR_QUALITY],
        fan_levels=list(range(1, 5)),
        device_alias='Core 400S',
        auto_preferences=[
            PurifierAutoPreference.DEFAULT,
            PurifierAutoPreference.EFFICIENT,
            PurifierAutoPreference.QUIET,
        ],
        model_display='Core 400S',
        model_name='Core 400S',
        setup_entry='Core400S',
    ),
    PurifierMap(
        class_name='VeSyncAirBypass',
        dev_types=['Core600S', 'LAP-C601S-WUS', 'LAP-C601S-WUSR', 'LAP-C601S-WEU'],
        modes=[PurifierModes.SLEEP, PurifierModes.MANUAL, PurifierModes.AUTO],
        features=[PurifierFeatures.AIR_QUALITY],
        auto_preferences=[
            PurifierAutoPreference.DEFAULT,
            PurifierAutoPreference.EFFICIENT,
            PurifierAutoPreference.QUIET,
        ],
        fan_levels=list(range(1, 5)),
        device_alias='Core 600S',
        model_display='Core 600S',
        model_name='Core 600S',
        setup_entry='Core600S',
    ),
    PurifierMap(
        class_name='VeSyncAir131',
        dev_types=['LV-PUR131S'],
        modes=[PurifierModes.SLEEP, PurifierModes.MANUAL, PurifierModes.AUTO],
        features=[PurifierFeatures.AIR_QUALITY],
        fan_levels=list(range(1, 4)),
        device_alias='LV-PUR131S',
        model_display='LV-PUR131S Series',
        model_name='LV131S',
        setup_entry='LV-PUR131S',
    ),
    PurifierMap(
        class_name='VeSyncAirRH131',
        dev_types=['LV-RH131S-WM', 'LV-RH131S'],
        modes=[PurifierModes.SLEEP, PurifierModes.MANUAL, PurifierModes.AUTO],
        features=[PurifierFeatures.AIR_QUALITY],
        fan_levels=list(range(1, 4)),
        device_alias='LV-RH131S',
        model_display='LV-RH131S Series',
        model_name='LV131S',
        setup_entry='LV-RH131S',
    ),
    PurifierMap(
        class_name='VeSyncAirBaseV2',
        dev_types=[
            'LAP-V102S-AASR',
            'LAP-V102S-WUS',
            'LAP-V102S-WEU',
            'LAP-V102S-AUSR',
            'LAP-V102S-WJP',
            'LAP-V102S-AJPR',
            'LAP-V102S-AEUR',
        ],
        modes=[
            PurifierModes.SLEEP,
            PurifierModes.MANUAL,
            PurifierModes.AUTO,
            PurifierModes.PET,
        ],
        features=[PurifierFeatures.AIR_QUALITY],
        fan_levels=list(range(1, 5)),
        device_alias='Vital 100S',
        auto_preferences=[
            PurifierAutoPreference.DEFAULT,
            PurifierAutoPreference.EFFICIENT,
            PurifierAutoPreference.QUIET,
        ],
        model_display='LAP-V102S Series',
        model_name='Vital 100S',
        setup_entry='LAP-V102S',
    ),
    PurifierMap(
        class_name='VeSyncAirBaseV2',
        dev_types=[
            'LAP-V201S-AASR',
            'LAP-V201S-WJP',
            'LAP-V201S-WEU',
            'LAP-V201S-WUS',
            'LAP-V201-AUSR',
            'LAP-V201S-AUSR',
            'LAP-V201S-AEUR',
        ],
        modes=[
            PurifierModes.SLEEP,
            PurifierModes.MANUAL,
            PurifierModes.AUTO,
            PurifierModes.PET,
        ],
        features=[PurifierFeatures.AIR_QUALITY, PurifierFeatures.LIGHT_DETECT],
        fan_levels=list(range(1, 5)),
        device_alias='Vital 200S',
        auto_preferences=[
            PurifierAutoPreference.DEFAULT,
            PurifierAutoPreference.EFFICIENT,
            PurifierAutoPreference.QUIET,
        ],
        model_display='LAP-V201S Series',
        model_name='Vital 200S',
        setup_entry='LAP-V201S',
    ),
    PurifierMap(
        class_name='VeSyncAirBaseV2',
        dev_types=[
            'LAP-EL551S-AUS',
            'LAP-EL551S-AEUR',
            'LAP-EL551S-WEU',
            'LAP-EL551S-WUS',
        ],
        modes=[
            PurifierModes.SLEEP,
            PurifierModes.MANUAL,
            PurifierModes.AUTO,
            PurifierModes.TURBO,
        ],
        features=[
            PurifierFeatures.AIR_QUALITY,
            PurifierFeatures.VENT_ANGLE,
            PurifierFeatures.LIGHT_DETECT,
        ],
        fan_levels=list(range(1, 4)),
        device_alias='Everest Air',
        auto_preferences=[
            PurifierAutoPreference.DEFAULT,
            PurifierAutoPreference.EFFICIENT,
            PurifierAutoPreference.QUIET,
        ],
        model_display='LAP-EL551S Series',
        model_name='Everest Air',
        setup_entry='EL551S',
    ),
    PurifierMap(
        class_name='VeSyncAirSprout',
        dev_types=[
            'LAP-B851S-WEU',
            'LAP-B851S-WNA',
            'LAP-B851S-AEUR',
            'LAP-B851S-AUS',
            'LAP-B851S-WUS',
            'LAP-BAY-MAX01S',
        ],
        modes=[
            PurifierModes.SLEEP,
            PurifierModes.MANUAL,
            PurifierModes.AUTO,
        ],
        features=[
            PurifierFeatures.AIR_QUALITY,
            PurifierFeatures.NIGHTLIGHT,
        ],
        fan_levels=list(range(1, 4)),
        device_alias='Sprout Air Purifier',
        auto_preferences=[
            PurifierAutoPreference.DEFAULT,
            PurifierAutoPreference.EFFICIENT,
            PurifierAutoPreference.QUIET,
        ],
        model_display='Sprout Air Series',
        model_name='Sprout Air',
        setup_entry='LAP-B851S-WUS',
    ),
]
"""List of ['PurifierMap'][pyvesync.device_map.PurifierMap] configuration
objects for purifier devices."""


fan_modules: list[FanMap] = [
    FanMap(
        class_name='VeSyncTowerFan',
        dev_types=['LTF-F422S-KEU', 'LTF-F422S-WUSR', 'LTF-F422S-WJP', 'LTF-F422S-WUS'],
        modes={
            FanModes.NORMAL: 'normal',
            FanModes.TURBO: 'turbo',
            FanModes.AUTO: 'auto',
            FanModes.SLEEP: 'advancedSleep',
        },
        set_mode_method='setTowerFanMode',
        features=[
            FanFeatures.OSCILLATION,
            FanFeatures.DISPLAYING_TYPE,
            FanFeatures.SOUND,
        ],
        fan_levels=list(range(1, 13)),
        device_alias='Tower Fan',
        sleep_preferences=[
            FanSleepPreference.DEFAULT,
            FanSleepPreference.ADVANCED,
            FanSleepPreference.TURBO,
            FanSleepPreference.EFFICIENT,
            FanSleepPreference.QUIET,
        ],  # Unknown sleep preferences, need to be verified
        model_display='LTF-F422S Series',
        model_name='Classic 42-Inch Tower Fan',
        setup_entry='LTF-F422S',
    ),
    FanMap(
        class_name='VeSyncPedestalFan',
        dev_types=['LPF-R432S-AEU', 'LPF-R432S-AUS'],
        modes={
            FanModes.NORMAL: 'normal',
            FanModes.TURBO: 'turbo',
            FanModes.ECO: 'eco',
            FanModes.SLEEP: 'advancedSleep',
        },
        setup_entry='LPF-R423S',
        features=[
            FanFeatures.SET_OSCILLATION_RANGE,
            FanFeatures.HORIZONTAL_OSCILLATION,
            FanFeatures.VERTICAL_OSCILLATION,
        ],
        fan_levels=list(range(1, 13)),
        set_mode_method='setFanMode',
        device_alias='Pedestal Fan',
        sleep_preferences=[
            FanSleepPreference.DEFAULT,
            FanSleepPreference.ADVANCED,
            FanSleepPreference.TURBO,
            FanSleepPreference.QUIET,
        ],  # Unknown sleep preferences, need to be verified
        model_display='LPF-R432S Pedestal Fan Series',
        model_name='Pedestal Fan',
    ),
]
"""List of ['FanMap'][pyvesync.device_map.FanMap] configuration
objects for fan devices."""


air_fryer_modules: list[AirFryerMap] = [
    AirFryerMap(
        class_name='VeSyncAirFryer158',
        module=vesynckitchen,
        dev_types=['CS137-AF/CS158-AF', 'CS158-AF', 'CS137-AF'],
        device_alias='Air Fryer',
        model_display='CS158/159/168/169-AF Series',
        model_name='Smart/Pro/Pro Gen 2 5.8 Qt. Air Fryer',
        setup_entry='CS137-AF/CS158-AF',
    )
]
"""List of ['AirFryerMap'][pyvesync.device_map.AirFryerMap] configuration
for air fryer devices."""


full_device_list = [
    *fan_modules,
    *purifier_modules,
    *humidifier_modules,
    *air_fryer_modules,
    *thermostat_modules,
]
"""List of all device configuration objects."""


def get_device_config(device_type: str) -> DeviceMapTemplate | None:
    """Get general device details from device type to create instance.

    Args:
        device_type (str): Device type to match from device list API call.

    Returns:
        DeviceMapTemplate | None: DeviceMapTemplate object or None if not found.
    """
    all_modules: list[T_MAPS] = [
        outlet_modules,
        switch_modules,
        bulb_modules,
        fan_modules,
        purifier_modules,
        humidifier_modules,
        air_fryer_modules,
        thermostat_modules,
    ]
    for module in chain(*all_modules):
        if device_type in module.dev_types:
            return module
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in chain(*all_modules):
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None


def get_fan(device_type: str) -> FanMap | None:
    """Get fan device details from device type.

    Args:
        device_type (str): Device type to match from device list API call.

    Returns:
        FanMap | None: FanMap object or None if not found.
    """
    for module in fan_modules:
        if device_type in module.dev_types:
            return module
    # Try to match with a more generic device type
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in fan_modules:
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None


def get_purifier(device_type: str) -> PurifierMap | None:
    """Get purifier device details from device type.

    Args:
        device_type (str): Device type to match from device list API call.

    Returns:
        PurifierMap | None: PurifierMap object or None if not found.
    """
    for module in purifier_modules:
        if device_type in module.dev_types:
            return module
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in purifier_modules:
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None


def get_humidifier(device_type: str) -> HumidifierMap | None:
    """Get humidifier device details from device type.

    Args:
        device_type (str): Device type to match from device list API call.

    Returns:
        HumidifierMap | None: HumidifierMap object or None if not found.
    """
    for module in humidifier_modules:
        if device_type in module.dev_types:
            return module
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in humidifier_modules:
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None


def get_outlet(device_type: str) -> OutletMap | None:
    """Get outlet device details from device type.

    Args:
        device_type (str): Device type to match from device list API call.

    Returns:
        OutletMap | None: OutletMap object or None if not found.
    """
    for module in outlet_modules:
        if device_type in module.dev_types:
            return module
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in outlet_modules:
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None


def get_switch(device_type: str) -> SwitchMap | None:
    """Get switch device details from device type.

    Args:
        device_type (str): Device type to match from device list API call.

    Returns:
        SwitchMap | None: SwitchMap object or None if not found.
    """
    for module in switch_modules:
        if device_type in module.dev_types:
            return module
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in switch_modules:
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None


def get_bulb(device_type: str) -> BulbMap | None:
    """Get bulb device details from device type.

    Args:
        device_type (str): Device type to match from device list API call.

    Returns:
        BulbMap | None: BulbMap object or None if not found.
    """
    for module in bulb_modules:
        if device_type in module.dev_types:
            return module
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in bulb_modules:
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None


def get_air_fryer(device_type: str) -> AirFryerMap | None:
    """Get air fryer device details from device type.

    Args:
        device_type (str): Device type to match from device list API call.

    Returns:
        AirFryerMap | None: AirFryerMap object or None if not found.
    """
    for module in air_fryer_modules:
        if device_type in module.dev_types:
            return module
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in air_fryer_modules:
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None


def get_thermostat(device_type: str) -> ThermostatMap | None:
    """Get the device map for a thermostat.

    Args:
        device_type (str): The device type to match.

    Returns:
        ThermostatMap | None: The matching thermostat map or None if not found.
    """
    for module in thermostat_modules:
        if device_type in module.dev_types:
            return module
    if device_type.count('-') > 1:
        device_type = '-'.join(device_type.split('-')[:-1])
        for module in thermostat_modules:
            if any(device_type.lower() in dev.lower() for dev in module.dev_types):
                return module
    return None



"""Data models for VeSync thermostats."""

from __future__ import annotations

from dataclasses import dataclass, field

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from pyvesync.const import ThermostatConst
from pyvesync.models.bypass_models import BypassV2InnerResult


@dataclass
class ResultThermostatDetails(BypassV2InnerResult):
    """Result model for thermostat details."""

    supportMode: list[int]
    workMode: int
    workStatus: int
    fanMode: int
    fanStatus: int
    tempUnit: str
    temperature: float
    humidity: int
    heatToTemp: int
    coolToTemp: int
    lockStatus: bool
    scheduleOrHold: int
    holdEndTime: int
    holdOption: int
    deadband: int
    ecoType: int
    alertStatus: int
    routines: list[ThermostatSimpleRoutine]
    routineRunningId: int | None = None


@dataclass
class ThermostatSimpleRoutine(DataClassORJSONMixin):
    """Thermostat routine model."""

    name: str
    routineId: int


@dataclass
class ThermostatRoutine(DataClassORJSONMixin):
    """Model for full thermostat routine."""

    name: str
    routineId: int
    type: int
    heatToTemp: int
    coolToTemp: int
    heatFanMode: int
    coolFanMode: int
    usuallyMask: int
    sensorIds: list[str]


@dataclass
class ThermostatMinorDetails(DataClassORJSONMixin):
    """Model for thermostat minor details."""

    mcu_version: str = field(metadata=field_options(alias='mcuVersion'))
    hvac_capacity: int = field(metadata=field_options(alias='hvacCapcity'))
    timestamp: int = field(metadata=field_options(alias='timeStamp'))
    time_zone: int = field(metadata=field_options(alias='timeZone'))
    offset_in_sec: int = field(metadata=field_options(alias='offsetInSec'))
    time_fmt: int = field(metadata=field_options(alias='timeFmt'))
    date_fmt: int = field(metadata=field_options(alias='dateFmt'))
    fan_delay_time: int = field(metadata=field_options(alias='fanDelayTime'))
    fan_circulation_time: int = field(metadata=field_options(alias='fanCirTime'))
    hvac_protect_time: int = field(metadata=field_options(alias='hvacProtecTime'))
    hvac_min_on_time: int = field(metadata=field_options(alias='hvacMinOnTime'))
    aux_min_on_time: int = field(metadata=field_options(alias='auxMinOnTime'))
    screen_brightness: int = field(metadata=field_options(alias='screenBrightness'))
    standby_timeout: int = field(metadata=field_options(alias='standbyTimeOut'))
    aux_low_temperature: int = field(metadata=field_options(alias='auxLowBalanceTemp'))
    aux_high_temperature: int = field(metadata=field_options(alias='auxHighBalanceTemp'))
    keytone: bool = field(metadata=field_options(alias='keyTone'))
    smart_schedule_enabled: bool = field(
        metadata=field_options(alias='smartScheduleEnabled')
    )
    time_to_temp_enabled: bool = field(metadata=field_options(alias='timeToTempEnabled'))
    early_on_enabled: bool = field(metadata=field_options(alias='earlyOnEnabled'))
    reminder_list: list[ThermostatReminder] = field(
        metadata=field_options(alias='reminderList')
    )
    alarm_list: list[ThermostatAlarm] = field(metadata=field_options(alias='alarmList'))


@dataclass
class ThermostatReminder(DataClassORJSONMixin):
    """Model for thermostat reminder."""

    code: int
    enabled: bool
    frequency: int
    code_name: str | None = None
    last_maintenance_time: int | None = field(
        default=None, metadata=field_options(alias='lastMaintenTime')
    )

    @classmethod
    def __post_deserialize__(  # type: ignore[override]
        cls, obj: ThermostatReminder
    ) -> ThermostatReminder:
        """Post-deserialization processing."""
        if isinstance(obj.code, int):
            obj.code_name = ThermostatConst.ReminderCode(obj.code).name
        return obj


@dataclass
class ThermostatAlarm(DataClassORJSONMixin):
    """Model for thermostat alarm."""

    code: int
    enabled: bool
    code_name: str | None = None
    aux_runtime_limit: int | None = field(
        default=None, metadata=field_options(alias='auxRunTimeLimit')
    )

    @classmethod
    def __post_deserialize__(  # type: ignore[override]
        cls, obj: ThermostatAlarm
    ) -> ThermostatAlarm:
        """Post-deserialization processing."""
        if obj.code is not None:
            obj.code_name = ThermostatConst.AlarmCode(obj.code).name
        return obj



from collections import namedtuple
import orjson
from pyvesync.utils.colors import RGB
from pyvesync.devices.vesyncbulb import (VeSyncBulbESL100, VeSyncBulbESL100CW,
                                 VeSyncBulbESL100MC, VeSyncBulbValcenoA19MC)
import call_json
import call_json_bulbs
from base_test_cases import TestBase

DEV_LIST = call_json.DeviceList.device_list_response('ESL100')

DEV_LIST_DETAIL = call_json.DeviceList.device_list_item('ESL100')

DEV_LIST_DETAIL_CW = call_json.DeviceList.device_list_item('ESL100CW')

DEV_LIST_CW = call_json.DeviceList.device_list_response('ESL100CW')

DEV_LIST_DETAIL_MC = call_json.DeviceList.device_list_item('ESL100MC')

DEV_LIST_MC = call_json.DeviceList.device_list_response('ESL100MC')

DEV_LIST_DETAIL_VALCENO = call_json.DeviceList.device_list_item('XYD0001')

DEV_LIST_VALCENO = call_json.DeviceList.device_list_response('XYD0001')

DEVICE_DETAILS = call_json_bulbs.DETAILS_RESPONSES['ESL100']

DEVICE_DETAILS_CW = call_json_bulbs.DETAILS_RESPONSES['ESL100CW']

DEFAULTS = call_json.TestDefaults

SUCCESS_RETURN = (orjson.dumps({'code': 0}), 200)


class TestVeSyncBulbESL100(TestBase):
    """Tests for VeSync dimmable bulb."""
    device_type = 'ESL100'

    def test_esl100_conf(self):
        """Tests that Wall Switch is instantiated properly."""
        self.mock_api.return_value = DEV_LIST
        self.run_in_loop(self.manager.get_devices)
        bulbs = self.manager.bulbs
        assert len(bulbs) == 1
        bulb = bulbs[0]
        assert isinstance(bulb, VeSyncBulbESL100)
        assert bulb.device_name == call_json.TestDefaults.name(self.device_type)
        assert bulb.device_type == self.device_type
        assert bulb.cid == call_json.TestDefaults.cid(self.device_type)
        assert bulb.uuid == call_json.TestDefaults.uuid(self.device_type)

    def test_esl100_details(self):
        """Test WS get_details()."""
        if callable(DEVICE_DETAILS):
            resp_dict, status = DEVICE_DETAILS()
        else:
            resp_dict, status = DEVICE_DETAILS
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        bulb = VeSyncBulbESL100(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(bulb.get_details)
        dev_details = bulb.details
        assert bulb.device_status == 'on'
        assert isinstance(dev_details, dict)
        assert bulb.connection_status == 'online'

    def test_esl100_no_details(self):
        """Test no device details for disconnected bulb."""
        self.mock_api.return_value = (orjson.dumps({'code': 5}), 200)
        bulb = VeSyncBulbESL100(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(bulb.update)
        assert len(self.caplog.records) == 1

    def test_esl100_onoff(self):
        """Test power toggle for ESL100 bulb."""
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        bulb = VeSyncBulbESL100(DEV_LIST_DETAIL, self.manager)
        assert self.run_in_loop(bulb.turn_off)
        assert self.run_in_loop(bulb.turn_on)

    def test_brightness(self):
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        bulb = VeSyncBulbESL100(DEV_LIST_DETAIL, self.manager)
        assert self.run_in_loop(bulb.set_brightness, 50)
        assert self.run_in_loop(bulb.turn_off)
        assert self.run_in_loop(bulb.set_brightness, 50)
        assert bulb.device_status == 'on'

    def test_invalid_brightness(self):
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        bulb = VeSyncBulbESL100(DEV_LIST_DETAIL, self.manager)
        assert self.run_in_loop(bulb.set_brightness, 5000)
        assert bulb.brightness == 100

    def test_features(self):
        bulb = VeSyncBulbESL100(DEV_LIST_DETAIL, self.manager)
        assert bulb.dimmable_feature
        assert not bulb.color_temp_feature
        assert not bulb.rgb_shift_feature


class TestVeSyncBulbESL100CW(TestBase):
    """Tests for VeSync dimmable bulb."""
    device_type = 'ESL100CW'

    def test_esl100cw_conf(self):
        """Tests that Wall Switch is instantiated properly."""
        if callable(DEV_LIST_CW):
            resp_dict, status = DEV_LIST_CW()
        else:
            resp_dict, status = DEV_LIST_CW
        if isinstance(resp_dict, str):
            resp_str = resp_dict.encode('utf-8')
        elif isinstance(resp_dict, dict):
            resp_str = orjson.dumps(resp_dict)
        elif isinstance(resp_dict, bytes):
            resp_str = resp_dict
        self.mock_api.return_value = resp_str, status
        self.run_in_loop(self.manager.get_devices)
        bulbs = self.manager.bulbs
        assert len(bulbs) == 1
        bulb = bulbs[0]
        assert isinstance(bulb, VeSyncBulbESL100CW)
        assert bulb.device_name == DEFAULTS.name(self.device_type)
        assert bulb.device_type == self.device_type
        assert bulb.cid == DEFAULTS.cid(self.device_type)
        assert bulb.uuid == DEFAULTS.uuid(self.device_type)

    def test_esl100cw_details(self):
        """Test WS get_details()."""
        if callable(DEVICE_DETAILS_CW):
            resp_dict, status = DEVICE_DETAILS_CW()
        else:
            resp_dict, status = DEVICE_DETAILS_CW
        if isinstance(resp_dict, str):
            resp_bytes = resp_dict.encode('utf-8')
        elif isinstance(resp_dict, dict):
            resp_bytes = orjson.dumps(resp_dict)
        elif isinstance(resp_dict, bytes):
            resp_bytes = resp_dict
        self.mock_api.return_value = resp_bytes, status
        bulb = VeSyncBulbESL100CW(DEV_LIST_DETAIL_CW, self.manager)
        self.run_in_loop(bulb.get_details)
        assert self.mock_api.r
        assert bulb.device_status == 'on'
        assert bulb.connection_status == 'online'

    def test_esl100cw_onoff(self):
        """Test power toggle for ESL100 bulb."""
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbESL100CW(DEV_LIST_DETAIL_CW, self.manager)
        assert self.run_in_loop(bulb.turn_off)
        assert bulb.device_status == 'off'
        assert self.run_in_loop(bulb.turn_on)
        assert bulb.device_status == 'on'

    def test_brightness(self):
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbESL100CW(DEV_LIST_DETAIL_CW, self.manager)
        assert self.run_in_loop(bulb.set_brightness, 50)
        assert bulb.brightness == 50
        assert self.run_in_loop(bulb.turn_off)
        assert self.run_in_loop(bulb.set_brightness, 50)
        assert bulb.device_status == 'on'

    def test_invalid_brightness(self):
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbESL100CW(DEV_LIST_DETAIL_CW, self.manager)
        assert self.run_in_loop(bulb.set_brightness, 5000)
        assert bulb.brightness == 100

    def test_color_temp(self):
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbESL100CW(DEV_LIST_DETAIL_CW, self.manager)
        assert self.run_in_loop(bulb.set_color_temp, 50)
        assert bulb.color_temp_pct == 50

    def test_features(self):
        bulb = VeSyncBulbESL100CW(DEV_LIST_DETAIL_CW, self.manager)
        assert bulb.dimmable_feature
        assert bulb.color_temp_feature
        assert not bulb.rgb_shift_feature


class TestVeSyncBulbESL100MC(TestBase):
    """Tests for VeSync dimmable bulb."""
    device_type = 'ESL100MC'

    def test_esl100mc_conf(self):
        """Tests that Wall Switch is instantiated properly."""
        self.mock_api.return_value = DEV_LIST_MC
        self.run_in_loop(self.manager.get_devices)
        bulbs = self.manager.bulbs
        assert len(bulbs) == 1
        bulb = bulbs[0]
        assert isinstance(bulb, VeSyncBulbESL100MC)
        assert bulb.device_name == DEFAULTS.name(self.device_type)
        assert bulb.device_type == self.device_type
        assert bulb.cid == DEFAULTS.cid(self.device_type)
        assert bulb.uuid == DEFAULTS.uuid(self.device_type)

    def test_esl100mc_details(self):
        """Test WS get_details()."""
        self.mock_api.return_value = call_json_bulbs.DETAILS_RESPONSES[self.device_type]
        bulb = VeSyncBulbESL100MC(DEV_LIST_DETAIL_MC, self.manager)
        self.run_in_loop(bulb.get_details)
        assert self.mock_api.r
        assert bulb.device_status == 'on'
        assert bulb.connection_status == 'online'

    def test_esl100mc_onoff(self):
        """Test power toggle for ESL100 bulb."""
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbESL100MC(DEV_LIST_DETAIL_MC, self.manager)
        assert self.run_in_loop(bulb.turn_off)
        assert bulb.device_status == 'off'
        assert self.run_in_loop(bulb.turn_on)
        assert bulb.device_status == 'on'

    def test_brightness(self):
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbESL100MC(DEV_LIST_DETAIL_MC, self.manager)
        assert self.run_in_loop(bulb.set_brightness, 50)
        assert bulb.brightness == 50
        assert self.run_in_loop(bulb.turn_off)
        assert self.run_in_loop(bulb.set_brightness, 50)
        assert bulb.device_status == 'on'

    def test_invalid_brightness(self):
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbESL100MC(DEV_LIST_DETAIL_MC, self.manager)
        assert self.run_in_loop(bulb.set_brightness, 5000)
        assert bulb.brightness == 100

    def test_color(self):
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbESL100MC(DEV_LIST_DETAIL_MC, self.manager)
        assert self.run_in_loop(bulb.set_rgb, 50, 100, 150)
        assert bulb.color_rgb == namedtuple('rgb', 'red green blue')(50, 100, 150)

    def test_features(self):
        bulb = VeSyncBulbESL100MC(DEV_LIST_DETAIL_MC, self.manager)
        assert bulb.dimmable_feature
        assert not bulb.color_temp_feature
        assert bulb.rgb_shift_feature


class TestVeSyncBulbValceno(TestBase):
    """Tests for VeSync Valceno bulb."""
    device_type = 'XYD0001'

    def test_valceno_conf(self):
        """Tests that Valceno is instantiated properly."""
        self.mock_api.return_value = DEV_LIST_VALCENO
        self.run_in_loop(self.manager.get_devices)
        bulbs = self.manager.bulbs
        assert len(bulbs) == 1
        bulb = bulbs[0]
        assert isinstance(bulb, VeSyncBulbValcenoA19MC)
        assert bulb.device_name == DEFAULTS.name(self.device_type)
        assert bulb.device_type == self.device_type
        assert bulb.cid == DEFAULTS.cid(self.device_type)
        assert bulb.uuid == DEFAULTS.uuid(self.device_type)

    def test_valceno_details(self):
        """Test Valceno get_details()."""
        self.mock_api.return_value = DEV_LIST_VALCENO
        bulb = VeSyncBulbValcenoA19MC(DEV_LIST_DETAIL_VALCENO, self.manager)
        self.run_in_loop(bulb.get_details)
        assert self.mock_api.r
        assert bulb.device_status == 'on'
        assert bulb.connection_status == 'online'

    def test_valceno_onoff(self):
        """Test power toggle for Valceno MC bulb."""
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbValcenoA19MC(DEV_LIST_DETAIL_VALCENO, self.manager)
        assert self.run_in_loop(bulb.turn_off)
        assert bulb.device_status == 'off'
        assert self.run_in_loop(bulb.turn_on)
        assert bulb.device_status == 'on'

    def test_brightness(self):
        """Test brightness on Valceno."""
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbValcenoA19MC(DEV_LIST_DETAIL_VALCENO, self.manager)
        assert self.run_in_loop(bulb.set_brightness, 50)
        assert bulb.brightness == 50
        assert self.run_in_loop(bulb.turn_off)
        assert self.run_in_loop(bulb.set_brightness, 50)
        assert bulb.device_status == 'on'

    def test_invalid_brightness(self):
        """Test invalid brightness on Valceno."""
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbValcenoA19MC(DEV_LIST_DETAIL_VALCENO, self.manager)
        assert not self.run_in_loop(bulb.set_brightness, 5000)

    def test_invalid_saturation(self):
        """Test invalid saturation on Valceno."""
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbValcenoA19MC(DEV_LIST_DETAIL_VALCENO, self.manager)
        assert self.run_in_loop(bulb.set_color_saturation, 100)
        body_dict = {
            "method": "setLightStatusV2",
            "source": "APP",
            "data":
                {
                    "force": 1,
                    "brightness": "",
                    "colorTemp": "",
                    "colorMode": "hsv",
                    "hue": "",
                    "saturation": 1000,
                    "value": ""
                }
        }
        mock_call = self.mock_api.call_args[1]['json_object']['payload']
        assert mock_call == body_dict

    def test_color(self):
        """Test set color on Valceno."""
        self.mock_api.return_value = (orjson.dumps({
            'code': 0,
            'msg': '',
            'result': {
                'code': 0,
                'result': {
                    "enabled": 'on',
                    "colorMode": 'hsv',
                    'brightness': 100,
                    'hue': 5833,
                    'saturation': 6700,
                    'value': 59
                }
            }
        }), 200)
        bulb = VeSyncBulbValcenoA19MC(DEV_LIST_DETAIL_VALCENO, self.manager)
        assert self.run_in_loop(bulb.set_rgb, 50, 100, 150)
        assert bulb.color_rgb == RGB(50, 100, 150)

    def test_hue(self):
        """Test hue on Valceno MC Bulb."""
        self.mock_api.return_value = SUCCESS_RETURN
        bulb = VeSyncBulbValcenoA19MC(DEV_LIST_DETAIL_VALCENO, self.manager)
        self.run_in_loop(bulb.set_color_hue, 230.5)
        body_dict = {
            "method": "setLightStatusV2",
            "source": "APP",
            "data":
                {
                    "force": 1,
                    "brightness": "",
                    "colorTemp": "",
                    "colorMode": "hsv",
                    "hue": 6403,
                    "saturation": "",
                    "value": ""
                }
        }
        mock_call = self.mock_api.call_args[1]['json_object']['payload']
        assert mock_call == body_dict

    def test_features(self):
        bulb = VeSyncBulbValcenoA19MC(DEV_LIST_DETAIL_VALCENO, self.manager)
        assert bulb.dimmable_feature
        assert bulb.color_temp_feature
        assert bulb.rgb_shift_feature



"""Utilities to handle Enums."""

from __future__ import annotations

import enum


class IntEnumMixin(enum.IntEnum):
    """Mixin class to handle missing enum values.

    Adds __missing__ method using the `extend_enum` function to
    return a new enum member with the name "UNKNOWN" and the
    missing value.
    """

    @classmethod
    def _missing_(cls: type[enum.IntEnum], value: object) -> enum.IntEnum:
        """Handle missing enum values by returning member with UNKNOWN name."""
        for member in cls:
            if member.value == value:
                return member
        unknown_enum_val = int.__new__(cls, value)  # type: ignore[call-overload]
        unknown_enum_val._name_ = 'UNKNOWN'
        unknown_enum_val._value_ = value  # type: ignore[assignment]
        unknown_enum_val.__objclass__ = cls.__class__  # type: ignore[assignment]
        return unknown_enum_val



"""Models for VeSync Bulb API responses and requests.

These models are used to serialize and deserialize the JSON responses from the
VeSync API. The models are used in the VeSync API class methods to provide
type hints and data validation.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Self, TypedDict

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin

from pyvesync.models.base_models import ResponseBaseModel, ResponseCodeModel
from pyvesync.models.bypass_models import BypassV2InnerResult, RequestBypassV1


@dataclass
class JSONCMD(DataClassORJSONMixin):
    """Tunable Bulb JSON CMD dict."""

    light: None | JSONCMDLight = None
    getLightStatus: None | str = None

    class Config(BaseConfig):
        """Configure the JSONCMD model."""

        omit_none = True


@dataclass
class JSONCMDLight(DataClassORJSONMixin):
    """Light JSON CMD dict."""

    action: str
    brightness: int | None = None
    colorTempe: int | None = None

    class Config(BaseConfig):
        """Configure the JSONCMDLight model."""

        omit_none = True


@dataclass
class RequestESL100Detail(RequestBypassV1):
    """Request model for Etekcity bulb details."""


@dataclass
class RequestESL100Status(RequestBypassV1):
    """Request model for Etekcity bulb details."""

    status: str


@dataclass
class RequestESL100Brightness(RequestBypassV1):
    """Request model for Etekcity bulb details."""

    status: str
    brightNess: int


@dataclass
class RequestESL100CWBase(RequestBypassV1):
    """Request model for ESL100CW bulb."""

    jsonCmd: JSONCMD


@dataclass
class ResponseESL100Detail(ResponseCodeModel):
    """Response model for Etekcity bulb details."""

    traceId: str
    code: int
    msg: str | None
    result: ResponseESL100DetailResult


@dataclass
class ResponseESL100DetailResult(ResponseBaseModel):
    """ESL100 Dimmable Bulb Device Detail Response."""

    deviceName: str | None
    name: str | None
    brightness: int | None = field(metadata=field_options(alias='brightNess'))
    activeTime: int | None
    deviceStatus: str = 'off'
    connectionStatus: str = 'offline'

    @classmethod
    def __post_deserialize__(  # type: ignore[override]
        cls, obj: Self
    ) -> Self:
        """Set values depending on color or white mode."""
        if obj.brightness is None:
            obj.brightness = 0
        if obj.activeTime is None:
            obj.activeTime = 0
        return obj


@dataclass
class ResponseESL100CWDetail(ResponseCodeModel):
    """Response model for Etekcity bulb details."""

    result: ResponseESL100CWDetailResult


@dataclass
class ResponseESL100CWLight(ResponseBaseModel):
    """ESL100CW Tunable Bulb Device Detail Response."""

    brightness: int | None
    action: str = 'on'
    colorTempe: int = 0


@dataclass
class ResponseESL100CWDetailResult(ResponseBaseModel):
    """Result model for ESL100CW Tunable bulb details."""

    light: ResponseESL100CWLight


@dataclass
class ResponseESL100MCStatus(ResponseCodeModel):
    """Response model for Etekcity bulb status."""

    result: ResponseESL100MCOuterResult


@dataclass
class ResponseESL100MCOuterResult:
    """ESL100MC Multi-Color Bulb Status Response."""

    traceId: str
    code: int
    result: ResponseESL100MCResult


@dataclass
class ResponseESL100MCResult(BypassV2InnerResult):
    """ESL100MC Multi-Color Bulb Status Response."""

    colorMode: str
    action: str
    brightness: int = 0
    red: int = 0
    green: int = 0
    blue: int = 0


@dataclass
class ResponseValcenoStatus(ResponseCodeModel):
    """Response model for Valceno bulb status."""

    result: ResponseValcenoOuterResult


@dataclass
class ResponseValcenoOuterResult(ResponseBaseModel):
    """Valceno Bulb Status Response."""

    result: ResponseValcenoStatusResult
    traceId: str = ''
    code: int = 0


@dataclass
class ResponseValcenoStatusResult(ResponseBaseModel):
    """Valceno Bulb Status Result."""

    colorMode: str = ''
    colorTemp: int = 0
    brightness: int = 0
    hue: int = 0
    saturation: int = 0
    value: int = 0
    enabled: str = 'off'


class ValcenoStatusPayload(TypedDict):
    """Typed Dict for setting Valceno bulb status."""

    colorMode: str
    colorTemp: int | str
    brightness: int | str
    hue: int | str
    saturation: int | str
    value: int | str
    force: int



"""Module to contain VeSync device instances.

Attributes:
    DeviceContainerInstance (DeviceContainer): Singleton instance of the DeviceContainer
        class. This is imported by the `vesync` module.

Classes:
    DeviceContainer: Container for VeSync device instances.
        This class should not be instantiated directly. Use the `DeviceContainerInstance`
        instead.
    _DeviceContainerBase: Base class for VeSync device
        container. Inherits from `MutableSet`.
"""

from __future__ import annotations

import logging
import re
from collections.abc import Iterator, MutableSet, Sequence
from typing import TYPE_CHECKING, TypeVar, cast

from pyvesync.base_devices.vesyncbasedevice import VeSyncBaseDevice
from pyvesync.const import ProductTypes
from pyvesync.device_map import get_device_config

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.base_devices.bulb_base import VeSyncBulb
    from pyvesync.base_devices.fan_base import VeSyncFanBase
    from pyvesync.base_devices.fryer_base import VeSyncFryer
    from pyvesync.base_devices.humidifier_base import VeSyncHumidifier
    from pyvesync.base_devices.outlet_base import VeSyncOutlet
    from pyvesync.base_devices.purifier_base import VeSyncPurifier
    from pyvesync.base_devices.switch_base import VeSyncSwitch
    from pyvesync.base_devices.thermostat_base import VeSyncThermostat
    from pyvesync.models.vesync_models import (
        ResponseDeviceDetailsModel,
        ResponseDeviceListModel,
    )


logger = logging.getLogger(__name__)

T = TypeVar('T')


def _clean_string(string: str) -> str:
    """Clean a string by removing non alphanumeric characters and making lowercase."""
    return re.sub(r'[^a0zA-Z0-9]', '', string)


class _DeviceContainerBase(MutableSet[VeSyncBaseDevice]):
    """Base class for VeSync device container.

    Inherits from `MutableSet` and defines the core MutableSet methods.
    """

    __slots__ = ('__weakref__', '_data')

    def __init__(
        self,
        sequence: Sequence[VeSyncBaseDevice] | None = None,
        /,
    ) -> None:
        """Initialize the DeviceContainer class."""
        self._data: set[VeSyncBaseDevice] = set()
        if isinstance(sequence, Sequence):
            self._data.update(sequence)

    def __iter__(self) -> Iterator[VeSyncBaseDevice]:
        """Iterate over the container."""
        return iter(self._data)

    def __len__(self) -> int:
        """Return the length of the container."""
        return len(self._data)

    def add(self, value: VeSyncBaseDevice) -> None:
        """Add a device to the container."""
        if value in self._data:
            logger.debug('Device already exists')
            return
        self._data.add(value)

    def remove(self, value: VeSyncBaseDevice) -> None:
        """Remove a device from the container."""
        self._data.remove(value)

    def clear(self) -> None:
        """Clear the container."""
        self._data.clear()

    def __contains__(self, value: object) -> bool:
        """Check if a device is in the container."""
        return value in self._data


class DeviceContainer(_DeviceContainerBase):
    """Container for VeSync device instances.

    /// admonition | Warning
    type: warning

    This class should not be instantiated directly. Use the `DeviceContainerInstance`
    instead.
    ///

    The `DeviceContainer` class is a container for VeSync device instances that
    inherits behavior from `MutableSet`. The container contains all VeSync devices
    and is used to store and manage device instances. The container is a singleton
    and is instantiated directly by the `DeviceContainerInstance` in the
    `device_container` module and imported as needed.

    Use the [`add_new_devices`][pyvesync.device_container.DeviceContainer.add_new_devices]
    class method to add devices from the device list model API response and
    `remove_stale_devices` to remove stale devices from the device list API response
    model. The device list response model is built in the
    [VeSync.get_devices()][pyvesync.vesync.VeSync.get_devices] method.

    Args:
        sequence (Sequence[VeSyncBaseDevice] | None): A sequence of device instances to
            initialize the container with. Typically this is not used directly,
            defaults to None.

    Attributes:
        _data (set[VeSyncBaseDevice]): The mutable set of devices in the container.
    """

    __slots__ = ()

    def __init__(
        self,
        sequence: Sequence[VeSyncBaseDevice] | None = None,
        /,
    ) -> None:
        """Initialize the DeviceContainer class."""
        super().__init__(sequence)

    def _build_device_instance(
        self, device: ResponseDeviceDetailsModel, manager: VeSync
    ) -> VeSyncBaseDevice | None:
        """Create a device from a single device model from the device list.

        Args:
            device (ResponseDeviceDetailsModel): The device details model from the
                device list response model.
            manager (VeSync): The VeSync instance to pass to the device instance

        Returns:
            VeSyncBaseDevice: The device instance created from the device list
                response model.

        Raises:
            VeSyncAPIResponseError: If the model is not an instance of
                `ResponseDeviceDetailsModel`.
        """
        device_features = get_device_config(device.deviceType)
        if device_features is None:
            logger.debug('Device type %s not found in device map', device.deviceType)
            return None
        dev_class = device_features.class_name
        dev_module = device_features.module
        device.productType = device_features.product_type
        # Import via string to avoid circular imports
        cls = getattr(dev_module, dev_class)
        return cls(device, manager, device_features)

    def add_device_from_model(
        self, device: ResponseDeviceDetailsModel, manager: VeSync
    ) -> None:
        """Add a single device from the device list response model.

        Args:
            device (ResponseDeviceDetailsModel): The device details model from the
                device list response model.
            manager (VeSync): The VeSync instance to pass to the device instance

        Raises:
            VeSyncAPIResponseError: If the model is not an instance of
                `ResponseDeviceDetailsModel`.
        """
        device_obj = self._build_device_instance(device, manager)
        if device_obj is not None:
            self.add(device_obj)
            logger.debug(
                'Added device %s - %s to container.',
                device_obj.device_name,
                device_obj.device_type,
            )

    def device_exists(self, cid: str, sub_device_no: int | None = None) -> bool:
        """Check if a device with the given cid & sub_dev_no exists.

        Args:
            cid (str): The cid of the device to check.
            sub_device_no (int): The sub_device_no of the device to check, defaults to 0
                for most devices.

        Returns:
            bool: True if the device exists, False otherwise.
        """
        new_hash = hash(cid + str(sub_device_no))
        return any(new_hash == hash(dev) for dev in self._data)

    def get_by_name(self, name: str, fuzzy: bool = False) -> VeSyncBaseDevice | None:
        """Forgiving method to get a device by name.

        Args:
            name (str): The name of the device to get.
            fuzzy (bool): Use a fuzzy match to find the device. Defaults to False.

        Returns:
            VeSyncBaseDevice | None: The device instance if found, None otherwise.

        Note:
            Fuzzy matching removes all non-alphanumeric characters and makes the string
            lowercase. If there are multiple devices with the same name, the first one
            found will be returned (a set is unordered).
        """
        for device in self._data:
            if (fuzzy and _clean_string(device.device_name) == _clean_string(name)) or (
                device.device_name == name
            ):
                return device
        return None

    def remove_by_cid(self, cid: str) -> bool:
        """Remove a device by cid.

        Args:
            cid (str): The cid of the device to remove.

        Returns:
            bool : True if the device was removed, False otherwise.
        """
        device_found: VeSyncBaseDevice | None = None
        for device in self._data:
            if device.cid == cid:
                device_found = device
                break
        if device_found is not None:
            self.remove(device_found)
            return True
        return False

    def discard(self, value: VeSyncBaseDevice) -> None:
        """Discard a device from the container.

        Args:
            value (VeSyncBaseDevice): The device to discard.
        """
        return self._data.discard(value)

    def remove_stale_devices(self, device_list_result: ResponseDeviceListModel) -> None:
        """Remove devices that are not in the provided list.

        Args:
            device_list_result (ResponseDeviceListModel): The device list response model
                from the VeSync API. This is generated by the `VeSync.get_devices()`
                method.
        """
        device_list = device_list_result.result.list
        new_hashes = [
            hash(device.cid + str(device.subDeviceNo)) for device in device_list
        ]
        remove_cids = []
        for device in self._data:
            if hash(device) not in new_hashes:
                logger.debug('Removing stale device %s', device.device_name)
                remove_cids.append(device.cid)
        for cid in remove_cids:
            self.remove_by_cid(cid)

    def add_new_devices(
        self, device_list_result: ResponseDeviceListModel, manager: VeSync
    ) -> None:
        """Add new devices to the container.

        Args:
            device_list_result (ResponseDeviceListModel): The device list response model
                from the VeSync API. This is generated by the `VeSync.get_devices()`
                method.
            manager (VeSync): The VeSync instance to pass to the device instance
        """
        device_list = device_list_result.result.list
        for device in device_list:
            if self.device_exists(device.cid, device.subDeviceNo) not in self._data:
                self.add_device_from_model(device, manager)

    @property
    def outlets(self) -> list[VeSyncOutlet]:
        """Return a list of devices that are outlets."""
        return [
            cast('VeSyncOutlet', device)
            for device in self
            if device.product_type == ProductTypes.OUTLET
        ]

    @property
    def switches(self) -> list[VeSyncSwitch]:
        """Return a list of devices that are switches."""
        return [
            cast('VeSyncSwitch', device)
            for device in self
            if device.product_type == ProductTypes.SWITCH
        ]

    @property
    def bulbs(self) -> list[VeSyncBulb]:
        """Return a list of devices that are lights."""
        return [
            cast('VeSyncBulb', device)
            for device in self
            if device.product_type == ProductTypes.BULB
        ]

    @property
    def air_purifiers(self) -> list[VeSyncPurifier]:
        """Return a list of devices that are air purifiers."""
        return [
            cast('VeSyncPurifier', device)
            for device in self
            if device.product_type == ProductTypes.PURIFIER
        ]

    @property
    def fans(self) -> list[VeSyncFanBase]:
        """Return a list of devices that are fans."""
        return [
            cast('VeSyncFanBase', device)
            for device in self
            if device.product_type == ProductTypes.FAN
        ]

    @property
    def humidifiers(self) -> list[VeSyncHumidifier]:
        """Return a list of devices that are humidifiers."""
        return [
            cast('VeSyncHumidifier', device)
            for device in self
            if device.product_type == ProductTypes.HUMIDIFIER
        ]

    @property
    def air_fryers(self) -> list[VeSyncFryer]:
        """Return a list of devices that are air fryers."""
        return [
            cast('VeSyncFryer', device)
            for device in self
            if device.product_type == ProductTypes.AIR_FRYER
        ]

    @property
    def thermostats(self) -> list[VeSyncThermostat]:
        """Return a list of devices that are thermostats."""
        return [
            cast('VeSyncThermostat', device)
            for device in self
            if device.product_type == ProductTypes.THERMOSTAT
        ]


DeviceContainerInstance = DeviceContainer()
"""Singleton instance of the DeviceContainer class.

This attribute should be imported by the `vesync` module and
not the `DeviceContainer` class directly.
"""



"""Request Models for Bypass V2 Endpoints.

API calls to bypassV2 endpoints have similar request structures. These models are used
to serialize and deserialize the JSON requests for the bypassV2 endpoints.
"""

from __future__ import annotations

from dataclasses import dataclass, field

import orjson
from mashumaro import pass_through
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin

from pyvesync.models.base_models import (
    BaseModelConfig,
    RequestBaseModel,
    ResponseBaseModel,
    ResponseCodeModel,
)


@dataclass
class RequestBypassV2(RequestBaseModel):
    """Bypass V2 Status Request Dict.

    This is the bypassV2 request model for API calls
    that use the `configModel` and `deviceId` fields.
    """

    acceptLanguage: str
    accountID: str
    appVersion: str
    cid: str
    configModule: str
    debugMode: bool
    method: str
    phoneBrand: str
    phoneOS: str
    traceId: str
    timeZone: str
    token: str
    userCountryCode: str
    deviceId: str
    configModel: str
    payload: BypassV2RequestPayload


@dataclass
class BypassV2RequestPayload(DataClassORJSONMixin):
    """Generic Bypass V2 Payload Request model."""

    data: dict
    method: str
    source: str = 'APP'
    subDeviceNo: int | None = None

    class Config(BaseModelConfig):
        """orjson config for dataclasses."""

        forbid_extra_keys = True
        orjson_options = orjson.OPT_NON_STR_KEYS
        omit_none = True


@dataclass
class RequestBypassV1(RequestBaseModel):
    """Bypass V1 Status Request Dict.

    This is the bypassV1 request model for API calls
    that use the `configModel` and `deviceId` fields.
    """

    acceptLanguage: str
    accountID: str
    appVersion: str
    cid: str
    configModule: str
    debugMode: bool
    method: str
    phoneBrand: str
    phoneOS: str
    traceId: str
    timeZone: str
    token: str
    userCountryCode: str
    uuid: str
    deviceId: str
    configModel: str

    class Config(BaseConfig):  # type: ignore[override]
        """Configure omit None value keys."""

        omit_none = True
        orjson_options = orjson.OPT_NON_STR_KEYS
        forbid_extra_keys = True


@dataclass
class ResponseBypassV1(ResponseCodeModel):
    """Bypass V1 Response Dict."""

    result: BypassV1Result | None = field(
        default=None,
        metadata={
            'serialize': pass_through,
            'deserialize': pass_through,
        },
    )


@dataclass
class BypassV1Result(DataClassORJSONMixin):
    """Bypass V1 Response Dict."""


@dataclass
class ResponseBypassV2(ResponseCodeModel):
    """Bypass V2 Response Dict.

    This is the bypassV2 response model for API calls
    that use the `configModel` and `deviceId` fields.
    """

    result: BypassV2OuterResult | None = None


@dataclass
class BypassV2InnerResult(DataClassORJSONMixin):
    """Inner Bypass V2 Result Data Model."""

    class Config(BaseModelConfig):
        """Configure the Outer Result model."""

        allow_deserialization_not_by_alias = True


@dataclass
class BypassV2OuterResult(DataClassORJSONMixin):
    """Bypass V2 Outer Result Data Model."""

    code: int
    result: BypassV2InnerResult | None = field(
        default=None,
        metadata={
            'serialize': pass_through,
            'deserialize': pass_through,
        },
    )

    class Config(BaseModelConfig):
        """Configure the Outer Result model."""

        allow_deserialization_not_by_alias = True


@dataclass
class BypassV2ResultError(DataClassORJSONMixin):
    """Bypass V2 Result Error Data Model."""

    msg: str


@dataclass
class ResultV2GetTimer(BypassV2InnerResult):
    """Inner result for Bypass V2 GetTimer method."""

    timers: list[TimerItemV2] | None = None


@dataclass
class ResultV2SetTimer(BypassV2InnerResult):
    """Result for Bypass V2 SetTimer method."""

    id: int


# Bypass V1 Timer Models


@dataclass
class TimeItemV1(ResponseBaseModel):
    """Data model for Bypass V1 Timers."""

    timerID: str
    counterTime: str
    action: str
    status: str
    resetTime: str
    uuid: str


@dataclass
class TimerItemV1(ResponseBaseModel):
    """Data model for Bypass V1 Timers."""

    timerID: str
    counterTimer: str
    action: str
    status: str
    resetTime: str


@dataclass
class TimerItemV2(ResponseBaseModel):
    """Data model for Bypass V2 Timers."""

    id: int
    remain: int
    action: str
    total: int


@dataclass
class TimerV2StartAct(ResponseBaseModel):
    """Action item for Timer V2 startAct array."""

    type: str
    act: int
    num: int


@dataclass
class TimerItemV2Alt(ResponseBaseModel):
    """Data model for Timer V2 API with startAct format."""

    id: int
    remain: int
    total: int
    startAct: list[TimerV2StartAct] = field(default_factory=list)


@dataclass
class ResultV2GetTimerV2(BypassV2InnerResult):
    """Inner result for GetTimerV2 method."""

    timers: list[TimerItemV2Alt] = field(default_factory=list)


@dataclass
class ResultV1SetTimer(BypassV1Result):
    """Result model for setting Bypass V1 API timer."""

    timerID: str
    conflictTimerIds: list[str] | None = None


@dataclass
class RequestV1ClearTimer(RequestBypassV1):
    """Request model for clearing Bypass V1 API outlet timer."""

    timerId: str
    status: str | None = None


@dataclass
class ResultV1GetTimerList(BypassV1Result):
    """Get timers result for v1 API."""

    timers: list[TimeItemV1] | list[TimerItemV1] | TimerItemV1 | None = None


@dataclass
class RequestV1SetTime(RequestBypassV1):
    """Request model for setting timer with counterTime."""

    counterTime: str
    action: str
    status: str | None = None
    switchNo: int | None = None


@dataclass
class RequestV1GetTimer(RequestBypassV1):
    """Request model for getting timers from v1 API."""

    switchNo: str | None = None


@dataclass
class RequestV1SetTimer(RequestBypassV1):
    """Request model for timer with counterTimer key.

    Attributes:
        counterTimer (str): The timer value in seconds.
        action (str): The action to perform (e.g., "on", "off").
        switchNo (int | None): The switch number for the timer.
    """

    counterTimer: str
    action: str
    switchNo: int | None = None


class TimerModels:
    """Class holding all common timer models.

    Attributes:
        ResultV2GetTimer (ResultV2GetTimer): Result model for Bypass V2 GetTimer method.
        ResultV2SetTimer (ResultV2SetTimer): Result model for Bypass V2 SetTimer method.
        ResultV1SetTimer (ResultV1SetTimer): Result model V1 API for setting timer.
        ResultV1GetTimer (ResultV1GetTimerList): Get timers result for v1 API.
        TimeItemV1 (TimeItemV1): Data model for Bypass V1 Timers.
        TimerItemV1 (TimerItemV1): Data model for Bypass V1 Timers.
        TimerItemV2 (TimerItemV2): Data model for Bypass V2 Timers.
        RequestV1ClearTimer (RequestV1ClearTimer): Model for deleting timer.
        RequestV1SetTimer (RequestV1SetTimer): Model for timer with counterTimer key.
        RequestV1GetTimer (RequestV1GetTimer): Model for getting timers from v1 API.
        RequestV1SetTime (RequestV1SetTime): Model for setting timer with counterTime key.
    """

    ResultV2GetTimer = ResultV2GetTimer
    ResultV2SetTimer = ResultV2SetTimer
    ResultV1SetTimer = ResultV1SetTimer
    ResultV1GetTimer = ResultV1GetTimerList
    TimeItemV1 = TimeItemV1
    TimerItemV1 = TimerItemV1
    TimerItemV2 = TimerItemV2
    RequestV1ClearTimer = RequestV1ClearTimer
    RequestV1SetTimer = RequestV1SetTimer
    RequestV1GetTimer = RequestV1GetTimer
    RequestV1SetTime = RequestV1SetTime



"""VeSync API Device Library."""

from __future__ import annotations

import asyncio
import logging
from dataclasses import MISSING, fields
from pathlib import Path
from typing import TYPE_CHECKING, Self

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientResponseError
from mashumaro.mixins.orjson import DataClassORJSONMixin

from pyvesync.auth import VeSyncAuth
from pyvesync.const import (
    DEFAULT_REGION,
    DEFAULT_TZ,
    MAX_API_REAUTH_RETRIES,
    REGION_API_MAP,
    STATUS_OK,
    ConnectionStatus,
    ProductTypes,
)
from pyvesync.device_container import DeviceContainer
from pyvesync.models.vesync_models import (
    FirmwareDeviceItemModel,
    RequestDeviceListModel,
    RequestFirmwareModel,
    ResponseDeviceListModel,
    ResponseFirmwareModel,
)
from pyvesync.utils.errors import (
    ErrorCodes,
    ErrorTypes,
    VeSyncAPIResponseError,
    VeSyncAPIStatusCodeError,
    VeSyncError,
    VeSyncServerError,
    VeSyncTokenError,
    raise_api_errors,
)
from pyvesync.utils.helpers import Helpers
from pyvesync.utils.logs import LibraryLogger

if TYPE_CHECKING:
    from pyvesync.base_devices import VeSyncBaseDevice

logger = logging.getLogger(__name__)


class VeSync:  # pylint: disable=function-redefined
    """VeSync Manager Class."""

    __slots__ = (
        '__weakref__',
        '_api_attempts',
        '_auth',
        '_close_session',
        '_debug',
        '_device_container',
        '_redact',
        '_verbose',
        'enabled',
        'in_process',
        'language',
        'session',
        'time_zone',
    )

    def __init__(
        self,
        username: str,
        password: str,
        country_code: str = DEFAULT_REGION,
        session: ClientSession | None = None,
        time_zone: str = DEFAULT_TZ,
        redact: bool = True,
    ) -> None:
        """Initialize VeSync Manager.

        This class is used as the manager for all VeSync objects, all methods and
        API calls are performed from this class. Time zone, debug and redact are
        optional. Time zone must be a string of an IANA time zone format. Once
        class is instantiated, call `await manager.login()` to log in to VeSync servers,
        which returns `True` if successful. Once logged in, call
        `await manager.get_devices()` to retrieve devices. Then `await `manager.update()`
        to update all devices or `await manager.devices[0].update()` to
        update a single device.

        Args:
            username (str): VeSync account username (usually email address)
            password (str): VeSync account password
            country_code (str): VeSync account country in ISO 3166 Alpha-2 format.
                By default, the account region is detected automatically at the login step
                If your account country is different from the default `US`,
                a second login attempt may be necessary - in this case
                you should specify the country directly to speed up the login process.
            session (ClientSession): aiohttp client session for
                API calls, by default None
            time_zone (str): Time zone for device from IANA database, by default
                DEFAULT_TZ. This is automatically set to the time zone of the
                VeSync account during login.
            redact (bool): Enable redaction of sensitive information, by default True.

        Attributes:
            session (ClientSession):  Client session for API calls
            devices (DeviceContainer): Container for all VeSync devices,
                has functionality of a mutable set. See
                [`DeviceContainer`][pyvesync.device_container.DeviceContainer] for
                more information
            auth (VeSyncAuth): Authentication manager
            time_zone (str): Time zone for VeSync account pulled from API
            enabled (bool): True if logged in to VeSync, False if not

        Note:
            This class is a context manager, use `async with VeSync() as manager:`
            to manage the session context. The session will be closed when exiting
            if no session is passed in.

            The `manager.devices` attribute is a DeviceContainer object that contains
            all VeSync devices. The `manager.devices` object has the functionality of
            a set, and can be iterated over to access devices. See :obj:`DeviceContainer`
            for more information.

            If using a context manager is not convenient, `manager.__aenter__()` and
            `manager.__aexit__()` can be called directly.

            Either username/password or token/account_id must be provided for
            authentication.

        See Also:
            :obj:`DeviceContainer`
                Container object to store VeSync devices
            :obj:`DeviceState`
                Object to store device state information
        """
        self.session = session
        self._api_attempts = 0
        self._close_session = False
        self.redact = redact
        self._verbose: bool = False
        self.time_zone: str = time_zone
        self.language: str = 'en'
        self.enabled = False
        self.in_process = False
        self._device_container: DeviceContainer = DeviceContainer()

        # Initialize authentication manager
        self._auth = VeSyncAuth(
            manager=self,
            username=username,
            password=password,
            country_code=country_code,
        )

    @property
    def devices(self) -> DeviceContainer:
        """Return VeSync device container.

        See Also:
            The pyvesync.device_container.DeviceContainer object
            for methods and properties.
        """
        return self._device_container

    @property
    def auth(self) -> VeSyncAuth:
        """Return VeSync authentication manager."""
        return self._auth

    @property
    def country_code(self) -> str:
        """Return country code."""
        return self._auth.country_code

    @property
    def current_region(self) -> str:
        """Return current region."""
        return self._auth.current_region

    @property
    def token(self) -> str:
        """Return authentication token.

        Returns:
            str: Authentication token.

        Raises:
            AttributeError: If token is not set.
        """
        return self._auth.token

    @property
    def account_id(self) -> str:
        """Return account ID.

        Returns:
            str: Account ID.

        Raises:
            AttributeError: If account ID is not set.
        """
        return self._auth.account_id

    @property
    def debug(self) -> bool:
        """Return debug flag."""
        return LibraryLogger.debug_enabled

    def check_debug(self) -> bool:
        """Check if debug logging is enabled - should be called minimally."""
        return LibraryLogger.check_debug()

    @property
    def redact(self) -> bool:
        """Return debug flag."""
        return self._redact

    @redact.setter
    def redact(self, new_flag: bool) -> None:
        """Set debug flag."""
        if new_flag:
            LibraryLogger.shouldredact = True
        elif new_flag is False:
            LibraryLogger.shouldredact = False
        self._redact = new_flag

    def output_credentials_json(self) -> str | None:
        """Output current authentication credentials as a JSON string."""
        return self.auth.output_credentials_json()

    def output_credentials_dict(self) -> dict[str, str] | None:
        """Output current authentication credentials as a dictionary."""
        return self.auth.output_credentials_dict()

    async def save_credentials(self, filename: str | Path | None) -> None:
        """Save authentication credentials to a file.

        Args:
            filename (str | Path | None): The name of the file to save credentials to.
                If None, no action is taken.
        """
        if filename is not None:
            await self.auth.save_credentials_to_file(filename)

    async def load_credentials_from_file(
        self, filename: str | Path | None = None
    ) -> bool:
        """Load authentication credentials from a file.

        Args:
            filename (str | Path | None): The name of the file to load credentials from.
                If None, no action is taken.

        Returns:
            bool: True if credentials were loaded successfully, False otherwise.
        """
        return await self.auth.load_credentials_from_file(filename)

    def set_credentials(
        self, token: str, account_id: str, country_code: str, region: str
    ) -> None:
        """Set authentication credentials.

        Args:
            token (str): Authentication token.
            account_id (str): Account ID.
            country_code (str): Country code in ISO 3166 Alpha-2 format.
            region (str): Current region code.
        """
        self._auth.set_credentials(token, account_id, country_code, region)

    def log_to_file(self, filename: str | Path, stdout: bool = True) -> None:
        """Log to file and enable debug logging.

        Args:
            filename (str | Path): The name of the file to log to.
            stdout (bool): Whether to also log to stdout, by default True.
        """
        LibraryLogger.configure_file_logging(filename, level=logging.DEBUG, stdout=stdout)
        logger.debug('Logging to file: %s', filename)

    def process_devices(self, dev_list_resp: ResponseDeviceListModel) -> bool:
        """Instantiate Device Objects.

        Internal method run by `get_devices()` to instantiate device objects.

        """
        current_device_count = len(self._device_container)
        if current_device_count > 0:
            self._device_container.remove_stale_devices(dev_list_resp)
        new_device_count = len(self._device_container)

        if new_device_count != current_device_count:
            logger.debug(
                'Removed %s devices', str(current_device_count - new_device_count)
            )

        current_device_count = new_device_count
        self._device_container.add_new_devices(dev_list_resp, self)
        new_device_count = len(self._device_container)
        if new_device_count != current_device_count:
            logger.debug('Added %s devices', str(new_device_count - current_device_count))
        return True

    async def get_devices(self) -> bool:
        """Return tuple listing outlets, switches, and fans of devices.

        This is also called by `VeSync.update()`

        Raises:
            VeSyncAPIResponseError: If API response is invalid.
            VeSyncServerError: If server returns an error.
        """
        self.in_process = True
        proc_return = False

        if not self.auth.is_authenticated or (
            not self.auth.token or not self.auth.account_id
        ):
            logger.info("Not logged in to VeSync, can't get devices")
            return False

        request_model = RequestDeviceListModel(
            token=self.auth.token, accountID=self.auth.account_id, timeZone=self.time_zone
        )
        logger.debug('Requesting device list from VeSync')
        response_dict, _ = await self.async_call_api(
            '/cloud/v1/deviceManaged/devices',
            'post',
            headers=Helpers.req_header_bypass(),
            json_object=request_model.to_dict(),
        )

        if response_dict is None:
            raise VeSyncAPIResponseError(
                'Error receiving response to device list request'
            )

        response = ResponseDeviceListModel.from_dict(response_dict)

        if response.code == 0:
            proc_return = self.process_devices(response)
        else:
            error_info = ErrorCodes.get_error_info(response.code)
            if response.msg is not None:
                error_info.message = f'{error_info.message} ({response.msg})'
            if error_info.error_type == ErrorTypes.SERVER_ERROR:
                raise VeSyncServerError(error_info.message)
            raise VeSyncAPIResponseError(
                'Error receiving response to device list request'
            )

        self.in_process = False

        return proc_return

    async def login(self) -> bool:  # pylint: disable=W9006 # pylint mult docstring raises
        """Log into VeSync server.

        Username and password are provided when class is instantiated.

        Returns:
            True if login successful, False otherwise

        Raises:
            VeSyncLoginError: If login fails, for example due to invalid username
                or password.
            VeSyncAPIResponseError: If API response is invalid.
            VeSyncServerError: If server returns an error.
        """
        self.enabled = False
        success = await self._auth.login()
        if success:
            self.enabled = True
        return success

    async def update(self) -> None:
        """Fetch updated information about devices and new device list.

        Pulls devices list from VeSync and instantiates any new devices. Devices
        are stored in the instance attributes `outlets`, `switches`, `fans`, and
        `bulbs`. The `_device_list` attribute is a dictionary of these attributes.
        """
        if not self.enabled:
            logger.error('Not logged in to VeSync')
            return
        await self.get_devices()

        await self.update_all_devices()

    async def update_all_devices(self) -> None:
        """Run `get_details()` for each device and update state."""
        logger.debug('Start updating the device details one by one')
        if len(self._device_container) == 0:
            logger.error('No devices to update')
            return
        update_tasks: list[asyncio.Task] = [
            asyncio.create_task(device.update()) for device in self._device_container
        ]
        done, _ = await asyncio.wait(update_tasks, return_when=asyncio.ALL_COMPLETED)
        for task in done:
            exc = task.exception()
            if exc is not None and isinstance(exc, VeSyncError):
                logger.error('Error updating device: %s', exc)

    async def __aenter__(self) -> Self:
        """Asynchronous context manager enter."""
        return self

    async def __aexit__(self, *exec_info: object) -> None:
        """Asynchronous context manager exit."""
        if self.session and self._close_session:
            logger.debug('Closing session, exiting context manager')
            await self.session.close()
            return
        logger.debug('Session not closed, exiting context manager')

    async def _reauthenticate(self) -> bool:
        """Re-authenticate using stored username and password.

        Returns:
            True if re-authentication successful, False otherwise
        """
        self.enabled = False
        self._api_attempts += 1
        if self._api_attempts >= MAX_API_REAUTH_RETRIES:
            logger.error('Max API re-authentication attempts reached')
            raise VeSyncTokenError
        success = await self.auth.reauthenticate()
        if success:
            self.enabled = True
            self._api_attempts = 0
            return True
        return await self.auth.reauthenticate()

    async def async_call_api(
        self,
        api: str,
        method: str,
        json_object: dict | None | DataClassORJSONMixin = None,
        headers: dict | None = None,
        device: VeSyncBaseDevice | None = None,
    ) -> tuple[dict | None, int]:
        """Make API calls by passing endpoint, header and body.

        api argument is appended to `API_BASE_URL`.
        Raises VeSyncRateLimitError if API returns a rate limit error.

        Args:
            api (str): Endpoint to call with `API_BASE_URL`.
            method (str): HTTP method to use.
            json_object (dict | RequestBaseModel): JSON object to send in body.
            headers (dict): Headers to send with request.
            device (VeSyncBaseDevice | None): Device making the request, if any.

        Returns:
            tuple[dict | None, int]: Response and status code. Attempts to parse
                response as JSON, if not possible returns None.

        Raises:
            VeSyncAPIStatusCodeError: If API returns an error status code.
            VeSyncRateLimitError: If API returns a rate limit error.
            VeSyncServerError: If API returns a server error.
            VeSyncTokenError: If API returns an authentication error.
            ClientResponseError: If API returns a client response error.

        Note:
            Future releases will require the `json_object` argument to be a dataclass,
            instead of dictionary.
        """
        self.check_debug()
        if self.session is None:
            self.session = ClientSession()
            self._close_session = True
        response = None
        status_code = None
        if isinstance(json_object, DataClassORJSONMixin):
            req_dict = json_object.to_dict()
        elif isinstance(json_object, dict):
            req_dict = json_object
        else:
            req_dict = None
        try:
            async with self.session.request(
                method,
                url=self._api_base_url_for_current_region() + api,
                json=req_dict,
                headers=headers,
                raise_for_status=False,
            ) as response:
                resp_bytes = await response.read()
                resp_status = response.status

                if resp_status != STATUS_OK:
                    LibraryLogger.log_api_status_error(logger, response=response)
                    raise VeSyncAPIStatusCodeError(str(resp_status))

                LibraryLogger.log_api_call(
                    logger,
                    response=response,
                    response_body=resp_bytes,
                    request_headers=headers,
                    request_body=req_dict,
                )
                resp_dict, status_code = await self._api_response_wrapper(
                    resp_bytes, resp_status, api, req_dict, device=device
                )

        except ClientResponseError as e:
            LibraryLogger.log_api_exception(logger, exception=e, request_body=req_dict)
            raise
        return resp_dict, status_code

    async def _api_response_wrapper(
        self,
        response_bytes: bytes | None,
        status_code: int,
        endpoint: str,
        request_body: dict | None,
        device: VeSyncBaseDevice | None = None,
    ) -> tuple[dict | None, int]:
        """Internal wrapper used by async_call_api."""
        resp_dict = LibraryLogger.try_json_loads(response_bytes)
        if not isinstance(resp_dict, dict):
            return None, status_code

        error_info = Helpers.parse_error_code(resp_dict)
        if error_info.error_type == ErrorTypes.TOKEN_ERROR:
            self.enabled = False
            if await self._reauthenticate():
                return await self.async_call_api(
                    endpoint, 'post', request_body, device=device
                )
            raise VeSyncTokenError(error_info.message)
        raise_api_errors(error_info)

        return resp_dict, status_code

    def _api_base_url_for_current_region(self) -> str:
        """Retrieve the API base url for the current region.

        At this point, only two different URLs exist: One for `EU` region
        (for all EU countries), and one for all others
        (currently `US`, `CA`, `MX`, `JP` - also used as a fallback).

        If `API_BASE_URL` is set, it will take precedence over the determined URL.
        """
        return REGION_API_MAP[self.current_region]

    def _update_fw_version(self, info_list: list[FirmwareDeviceItemModel]) -> bool:
        """Update device firmware versions from API response."""
        if not info_list:
            logger.info('No devices found in firmware response')
            return False
        update_dict = {}
        for device in info_list:
            if not device.firmUpdateInfos:
                if device.code != 0:
                    logger.info(
                        'Device %s has error code %s with message: %s',
                        device.deviceName,
                        device.code,
                        device.msg,
                    )
                else:
                    logger.debug(
                        'Device %s has no firmware updates available', device.deviceName
                    )
                continue
            for update_info in device.firmUpdateInfos:
                update_dict[device.deviceCid] = (
                    update_info.currentVersion,
                    update_info.latestVersion,
                )
                if update_info.isMainFw is True:
                    break
        for device_obj in self._device_container:
            if device_obj.cid in update_dict:
                device_obj.latest_firm_version = update_dict[device_obj.cid][1]
                device_obj.current_firm_version = update_dict[device_obj.cid][0]
        return True

    async def check_firmware(self) -> bool:
        """Check for firmware updates for all devices.

        This method will check for firmware updates for all devices in the
        device container. It will call the `get_firmware_update()` method on
        each device and log the results.
        """
        if len(self.devices) == 0:
            logger.warning('No devices to check for firmware updates')
            return False
        body_fields = tuple(
            field.name
            for field in fields(RequestFirmwareModel)
            if field.default_factory is MISSING and field.default is MISSING
        )
        body = Helpers.get_manager_attributes(self, body_fields)

        # Add cid list of devices that are online and not an air fryer
        body['cidList'] = [
            device.cid
            for device in self.devices
            if (
                device.product_type != ProductTypes.AIR_FRYER
                and device.state.connection_status == ConnectionStatus.ONLINE
            )
        ]

        if len(body['cidList']) == 0:
            logger.warning(
                'No online devices (excluding air fryers) to check firmware for'
            )
            return False

        resp_dict, _ = await self.async_call_api(
            '/cloud/v2/deviceManaged/getFirmwareUpdateInfoList',
            'post',
            json_object=RequestFirmwareModel(**body),
        )
        if resp_dict is None:
            raise VeSyncAPIResponseError(
                'Error receiving response to firmware update request'
            )
        resp_model = ResponseFirmwareModel.from_dict(resp_dict)
        if resp_model.code != 0:
            error_info = ErrorCodes.get_error_info(resp_model.code)
            resp_message = resp_model.msg
            if resp_message is not None:
                error_info.message = f'{error_info.message} ({resp_message})'
            logger.warning('Error in firmware update response: %s', error_info.message)
            return False
        info_list = resp_model.result.cidFwInfoList
        return self._update_fw_version(info_list)



"""Utility hooks for mkDocs."""
import re
import urllib.request


def replacer(match: re.Match) -> str:
    """Find and replace github links with code snippets.

    This function parses a Github link and retrieves the specified
    lines of code from the file, formatting it as a markdown code block.
    """
    filename = f'{match.group(3)}.{match.group(4)}'
    url = (
        f'https://raw.githubusercontent.com/{match.group(1)}/{match.group(2)}/{filename}'
    )
    code = urllib.request.urlopen(url).read().decode('utf-8')  # noqa: S310
    extension = 'js' if match.group(4) == 'vue' else match.group(4)
    return '\n'.join(
        [f'``` {extension} title="{filename}"']
        + code.split('\n')[int(match.group(5)) - 1: int(match.group(6))]
        + ['```', f'View at [GitHub]({match.group(0)})']
    )


def on_page_markdown(markdown: str, **kwargs) -> str:  # noqa: ARG001, ANN003
    """Replace GitHub links with code snippets."""
    return re.sub(
        re.compile(
            r'^https://github.com/([\w/\-]+)/blob/([0-9a-f]+)/([\w\d\-/\.]+)\.(\w+)#L(\d+)-L(\d+)$',
            re.MULTILINE,
        ),
        replacer,
        markdown,
    )



"""
Air Purifier and Humidifier Device API Responses

FANS variable is a list of setup_entry's from the device_map

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a Default Dictionary of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'DEV_TYPE': defaultdict(
        lambda: {"code": 0, "msg": "success"}
    )
}

# For a function to handle the response
def status_response(request_body=None):
    # do work with request_body
    return request_body, 200

METHOD_RESPONSES['DEV_TYPE']['set_status'] = status_response

# To change the default value for a device type

METHOD_RESPONSES['DEVTYPE'].default_factory = lambda: {"code": 0, "msg": "success"}
"""
from typing import Any
from copy import deepcopy

from pyvesync.device_map import humidifier_modules
from pyvesync.const import HumidifierModes, DeviceStatus, ConnectionStatus, DryingModes
from defaults import TestDefaults, FunctionResponses, build_bypass_v2_response, FunctionResponsesV2

HUMIDIFIERS = [m.setup_entry for m in humidifier_modules]
HUMIDIFIERS_NUM = len(HUMIDIFIERS)


class HumidifierDefaults:
    device_status = DeviceStatus.ON
    connection_status = ConnectionStatus.ONLINE
    humidifier_mode = HumidifierModes.MANUAL
    child_lock_switch = DeviceStatus.OFF
    temperature = 70 * 10
    humidity = 50
    wick_life_percent = 85
    auto_mode_preference = 1
    target_humidity = 60
    mist_level = 3
    virtual_mist_level = 3
    water_lacks = False
    water_tank_lifted = False
    auto_stop = False
    auto_stop_reached = False
    display = True
    display_config = True
    humidity_high = False
    hepa_filter_pct = 78
    # Warm Mist Defaults
    warm_mist_level = 1
    warm_mist_enabled = True
    # Nightlight Default
    nightlight_status = DeviceStatus.OFF
    nightlight_brightness = 0
    nightlight_color_temperature = 4000
    # Drying Defaults
    drying_level = 1
    drying_state = DryingModes.DONE
    water_shortage_drying_switch = DeviceStatus.ON
    drying_mode_switch = DeviceStatus.ON
    # Breating Lamp Defaults
    breathing_lamp_switch = DeviceStatus.OFF
    breathing_lamp_color_temperature = 4000
    breathing_lamp_time_interval = 10
    breathing_lamp_brightness_start = 0
    breathing_lamp_brightness_end = 50


HUMIDIFIER_DETAILS: dict[str, Any] = {
    "Classic300S": {
        "enabled": bool(HumidifierDefaults.device_status),
        "humidity": HumidifierDefaults.humidity,
        "mist_virtual_level": HumidifierDefaults.virtual_mist_level,
        "mist_level": HumidifierDefaults.mist_level,
        "mode": HumidifierDefaults.humidifier_mode.value,
        "water_lacks": HumidifierDefaults.water_lacks,
        "humidity_high": HumidifierDefaults.humidity_high,
        "water_tank_lifted": HumidifierDefaults.water_tank_lifted,
        "display": HumidifierDefaults.display,
        "automatic_stop_reach_target": HumidifierDefaults.auto_stop_reached,
        "night_light_brightness": HumidifierDefaults.nightlight_brightness,
        "configuration": {
            "auto_target_humidity": HumidifierDefaults.target_humidity,
            "display": HumidifierDefaults.display_config,
            "automatic_stop": HumidifierDefaults.auto_stop,
        },
    },
    "Classic200S": {
        "enabled": bool(HumidifierDefaults.device_status),
        "mist_virtual_level": HumidifierDefaults.virtual_mist_level,
        "mist_level": HumidifierDefaults.mist_level,
        "mode": HumidifierDefaults.humidifier_mode.value,
        "water_lacks": HumidifierDefaults.water_lacks,
        "water_tank_lifted": HumidifierDefaults.water_tank_lifted,
        "humidity": HumidifierDefaults.humidity,
        "humidity_high": HumidifierDefaults.humidity_high,
        "indicator_light_switch": HumidifierDefaults.display,
        "automatic_stop_reach_target": HumidifierDefaults.auto_stop_reached,
        "configuration": {
            "auto_target_humidity": HumidifierDefaults.target_humidity,
            "indicator_light_switch": HumidifierDefaults.display_config,
            "automatic_stop": HumidifierDefaults.auto_stop,
        },
    },
    "Dual200S": {
        "enabled": bool(HumidifierDefaults.device_status),
        "humidity": HumidifierDefaults.humidity,
        "mist_virtual_level": HumidifierDefaults.virtual_mist_level,
        "mist_level": HumidifierDefaults.mist_level,
        "mode": HumidifierDefaults.humidifier_mode.value,
        "water_lacks": HumidifierDefaults.water_lacks,
        "humidity_high": HumidifierDefaults.humidity_high,
        "water_tank_lifted": HumidifierDefaults.water_tank_lifted,
        "display": HumidifierDefaults.display,
        "automatic_stop_reach_target": HumidifierDefaults.auto_stop_reached,
        "night_light_brightness": HumidifierDefaults.nightlight_brightness,
        "configuration": {
            "auto_target_humidity": HumidifierDefaults.target_humidity,
            "display": HumidifierDefaults.display_config,
            "automatic_stop": HumidifierDefaults.auto_stop,
        },
    },
    "LUH-A602S-WUS": {  # LV600S
        "enabled": bool(HumidifierDefaults.device_status),
        "humidity": HumidifierDefaults.humidity,
        "mist_virtual_level": HumidifierDefaults.virtual_mist_level,
        "mist_level": HumidifierDefaults.mist_level,
        "mode": HumidifierDefaults.humidifier_mode.value,
        "water_lacks": HumidifierDefaults.water_lacks,
        "humidity_high": HumidifierDefaults.humidity_high,
        "water_tank_lifted": HumidifierDefaults.water_tank_lifted,
        "display": HumidifierDefaults.display,
        "automatic_stop_reach_target": HumidifierDefaults.auto_stop_reached,
        "night_light_brightness": HumidifierDefaults.nightlight_brightness,
        "warm_mist_level": HumidifierDefaults.warm_mist_level,
        "warm_mist_enabled": HumidifierDefaults.warm_mist_enabled,
        "configuration": {
            "auto_target_humidity": HumidifierDefaults.target_humidity,
            "display": HumidifierDefaults.display_config,
            "automatic_stop": HumidifierDefaults.auto_stop,
        },
    },
    "LUH-O451S-WUS": {
        "enabled": bool(HumidifierDefaults.device_status),
        "mist_virtual_level": HumidifierDefaults.virtual_mist_level,
        "mist_level": HumidifierDefaults.mist_level,
        "mode": HumidifierDefaults.humidifier_mode.value,
        "water_lacks": HumidifierDefaults.water_lacks,
        "water_tank_lifted": HumidifierDefaults.water_tank_lifted,
        "humidity": HumidifierDefaults.humidity,
        "humidity_high": HumidifierDefaults.humidity_high,
        "display": HumidifierDefaults.display,
        "warm_enabled": HumidifierDefaults.warm_mist_enabled,
        "warm_level": HumidifierDefaults.warm_mist_level,
        "automatic_stop_reach_target": HumidifierDefaults.auto_stop_reached,
        "configuration": {
            "auto_target_humidity": HumidifierDefaults.target_humidity,
            "display": HumidifierDefaults.display_config,
            "automatic_stop": HumidifierDefaults.auto_stop,
        },
        "extension": {"schedule_count": 0, "timer_remain": 0},
    },
    "LUH-O451S-WEU": {
        "enabled": bool(HumidifierDefaults.device_status),
        "mist_virtual_level": HumidifierDefaults.virtual_mist_level,
        "mist_level": HumidifierDefaults.mist_level,
        "mode": HumidifierDefaults.humidifier_mode.value,
        "water_lacks": HumidifierDefaults.water_lacks,
        "water_tank_lifted": HumidifierDefaults.water_tank_lifted,
        "humidity": HumidifierDefaults.humidity,
        "humidity_high": HumidifierDefaults.humidity_high,
        "display": HumidifierDefaults.display,
        "warm_enabled": HumidifierDefaults.warm_mist_enabled,
        "warm_level": HumidifierDefaults.warm_mist_level,
        "automatic_stop_reach_target": HumidifierDefaults.auto_stop_reached,
        "configuration": {
            "auto_target_humidity": HumidifierDefaults.target_humidity,
            "display": HumidifierDefaults.display_config,
            "automatic_stop": HumidifierDefaults.auto_stop,
        },
        "extension": {"schedule_count": 0, "timer_remain": 0},
    },
    "LUH-M101-WUS": {
        "powerSwitch": int(HumidifierDefaults.device_status),
        "humidity": int(HumidifierDefaults.humidity),
        "targetHumidity": int(HumidifierDefaults.target_humidity),
        "virtualLevel": int(HumidifierDefaults.virtual_mist_level),
        "mistLevel": int(HumidifierDefaults.mist_level),
        "workMode": HumidifierDefaults.humidifier_mode.value,
        "waterLacksState": int(HumidifierDefaults.water_lacks),
        "waterTankLifted": int(HumidifierDefaults.water_tank_lifted),
        "autoStopSwitch": int(HumidifierDefaults.auto_stop),
        "autoStopState": int(HumidifierDefaults.auto_stop_reached),
        "screenSwitch": int(HumidifierDefaults.display_config),
        "screenState": int(HumidifierDefaults.display),
        "scheduleCount": 0,
        "timerRemain": 0,
        "errorCode": 0,
    },
    "LUH-M101S-WEUR": {
        "powerSwitch": int(DeviceStatus.ON),
        "humidity": HumidifierDefaults.humidity,
        "targetHumidity": HumidifierDefaults.target_humidity,
        "virtualLevel": HumidifierDefaults.virtual_mist_level,
        "mistLevel": HumidifierDefaults.mist_level,
        "workMode": HumidifierDefaults.humidifier_mode.value,
        "waterLacksState": int(HumidifierDefaults.water_lacks),
        "waterTankLifted": int(HumidifierDefaults.water_tank_lifted),
        "autoStopSwitch": int(HumidifierDefaults.auto_stop),
        "autoStopState": int(HumidifierDefaults.auto_stop_reached),
        "screenSwitch": int(HumidifierDefaults.display_config),
        "screenState": int(HumidifierDefaults.display),
        "nightLight": {
            "nightLightSwitch": int(HumidifierDefaults.nightlight_status),
            "brightness": HumidifierDefaults.nightlight_brightness,
        },
        "scheduleCount": 0,
        "timerRemain": 0,
        "errorCode": 0,
    },
    "LEH-S601S": {
        "powerSwitch": int(DeviceStatus.ON),
        "humidity": HumidifierDefaults.humidity,
        "targetHumidity": HumidifierDefaults.target_humidity,
        "virtualLevel": HumidifierDefaults.virtual_mist_level,
        "mistLevel": HumidifierDefaults.mist_level,
        "workMode": HumidifierDefaults.humidifier_mode.value,
        "waterLacksState": int(HumidifierDefaults.water_lacks),
        "waterTankLifted": int(HumidifierDefaults.water_tank_lifted),
        "autoStopSwitch": int(HumidifierDefaults.auto_stop),
        "autoStopState": int(HumidifierDefaults.auto_stop_reached),
        "screenSwitch": int(HumidifierDefaults.display_config),
        "screenState": int(HumidifierDefaults.display),
        "scheduleCount": 0,
        "timerRemain": 0,
        "errorCode": 0,
        "dryingMode": {
            "dryingLevel": HumidifierDefaults.drying_level,
            "autoDryingSwitch": int(HumidifierDefaults.drying_mode_switch),
            "dryingState": int(HumidifierDefaults.drying_state),
            "dryingRemain": 7200,
        },
        "autoPreference": HumidifierDefaults.auto_mode_preference,
        "childLockSwitch": int(HumidifierDefaults.child_lock_switch),
        "filterLifePercent": HumidifierDefaults.wick_life_percent,
        "temperature": HumidifierDefaults.temperature,
    },
    "LEH-B381S": {  # Sprout Humidifier
        "powerSwitch": int(HumidifierDefaults.device_status),
        "humidity": HumidifierDefaults.humidity,
        "targetHumidity": HumidifierDefaults.target_humidity,
        "virtualLevel": HumidifierDefaults.virtual_mist_level,
        "mistLevel": HumidifierDefaults.mist_level,
        "workMode": HumidifierDefaults.humidifier_mode.value,
        "waterLacksState": int(HumidifierDefaults.water_lacks),
        "waterTankLifted": int(HumidifierDefaults.water_tank_lifted),
        "autoStopSwitch": int(HumidifierDefaults.auto_stop),
        "autoStopState": int(HumidifierDefaults.auto_stop_reached),
        "screenSwitch": int(HumidifierDefaults.display_config),
        "screenState": int(HumidifierDefaults.display),
        "autoModePreference": HumidifierDefaults.auto_mode_preference,
        "autoPreference": 4,  # Unknown value - autoModePreference shows the set mode
        "dryingMode": {
            "dryingLevel": HumidifierDefaults.drying_level,
            "autoDryingSwitch": HumidifierDefaults.drying_mode_switch,
            "dryingState": int(HumidifierDefaults.drying_state),
            "dryingRemain": 0,
        },
        "waterShortageDryingSwitch": int(HumidifierDefaults.water_shortage_drying_switch),
        "childLockSwitch": int(HumidifierDefaults.child_lock_switch),
        "filterLifePercent": HumidifierDefaults.wick_life_percent,
        "hepaFilterLifePercent": HumidifierDefaults.hepa_filter_pct,
        "sensorContent": {
            "selfSensorMac": "a4:c1:38:83:16:9b",
            "selfSenorBattery": 100,
            "selfSensorStatus": "on",
        },
        "temperature": HumidifierDefaults.temperature,
        "nightLight": {
            "nightLightSwitch": int(HumidifierDefaults.nightlight_status),
            "brightness": HumidifierDefaults.nightlight_brightness,
            "colorTemperature": HumidifierDefaults.nightlight_color_temperature,
        },
        "breathingLamp": {
            "breathingLampSwitch": int(HumidifierDefaults.breathing_lamp_switch),
            "colorTemperature": HumidifierDefaults.breathing_lamp_color_temperature,
            "timeInterval": HumidifierDefaults.breathing_lamp_time_interval,
            "brightnessStart": HumidifierDefaults.breathing_lamp_brightness_start,
            "brightnessEnd": HumidifierDefaults.breathing_lamp_brightness_end,
        },
        "lampSwitch": int(HumidifierDefaults.breathing_lamp_switch),
        "lampType": 1,
        "dumpedState": 0,
        "guardingInfo": {"guarding": 1, "remainTS": 0},
        "roomSize": 242,
        "supportLampAct": 1,
        "lastDryingCompletedTime": 1764685402,
        "afterDryLastHumidityTime": 1764651910,
    },
    "LUH-A603S-WUS": {  # LV600S with warm mist
        "powerSwitch": int(DeviceStatus.ON),
        "humidity": HumidifierDefaults.humidity,
        "targetHumidity": HumidifierDefaults.target_humidity,
        "virtualLevel": HumidifierDefaults.virtual_mist_level,
        "mistLevel": HumidifierDefaults.mist_level,
        "workMode": HumidifierDefaults.humidifier_mode.value,
        "waterLacksState": int(HumidifierDefaults.water_lacks),
        "waterTankLifted": int(HumidifierDefaults.water_tank_lifted),
        "autoStopSwitch": int(HumidifierDefaults.auto_stop),
        "autoStopState": int(HumidifierDefaults.auto_stop_reached),
        "screenSwitch": int(HumidifierDefaults.display_config),
        "screenState": int(HumidifierDefaults.display),
        "scheduleCount": 0,
        "timerRemain": 0,
        "errorCode": 0,
        "totalWorkTime": 0,
        "warmPower": HumidifierDefaults.warm_mist_enabled,
        "warmLevel": HumidifierDefaults.warm_mist_level,
    },
}
"""This dictionary contains the details response for each humidifier.

It stores the innermost result that is passed to the DETAILS_RESPONSE variable where
the full API response is built."""


DETAILS_RESPONSES = {
    "Classic300S": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["Classic300S"]),
    "Classic200S": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["Classic200S"]),
    "Dual200S": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["Dual200S"]),
    "LUH-A602S-WUS": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["LUH-A602S-WUS"]),
    "LUH-A603S-WUS": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["LUH-A603S-WUS"]),
    "LUH-O451S-WUS": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["LUH-O451S-WUS"]),
    "LUH-O451S-WEU": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["LUH-O451S-WEU"]),
    "LUH-M101S-WEUR": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["LUH-M101S-WEUR"]),
    "LUH-M101S-WUS": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["LUH-M101-WUS"]),
    "LEH-S601S": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["LEH-S601S"]),
    "LEH-B381S": build_bypass_v2_response(inner_result=HUMIDIFIER_DETAILS["LEH-B381S"]),
}


FunctionResponses.default_factory = lambda: (
    {
        "traceId": TestDefaults.trace_id,
        "code": 0,
        "msg": "request success",
        "result": {"traceId": TestDefaults.trace_id, "code": 0},
    },
    200,
)

METHOD_RESPONSES = {
    "Classic300S": deepcopy(FunctionResponsesV2),
    "Classic200S": deepcopy(FunctionResponsesV2),
    "Dual200S": deepcopy(FunctionResponsesV2),
    "LUH-A602S-WUS": deepcopy(FunctionResponsesV2),
    "LUH-A603S-WUS": deepcopy(FunctionResponsesV2),
    "LUH-O451S-WUS": deepcopy(FunctionResponsesV2),
    "LUH-O451S-WEU": deepcopy(FunctionResponsesV2),
    "LUH-M101S-WEUR": deepcopy(FunctionResponsesV2),
    "LUH-M101S-WUS": deepcopy(FunctionResponsesV2),
    "LEH-S601S": deepcopy(FunctionResponsesV2),
    "LEH-B381S": deepcopy(FunctionResponsesV2),
}


# Add responses for methods with different responses than the default

# Timer Responses

# for k in AIR_MODELS:
#     METHOD_RESPONSES[k]["set_timer"] = (INNER_RESULT({"id": 1}), 200)
#     METHOD_RESPONSES[k]["get_timer"] = (
#         INNER_RESULT({"id": 1, "remain": 100, "total": 100, "action": "off"}),
#         200,
#     )
# FAN_TIMER = helpers.Timer(100, "off")



"""Common base class for tests and Default values.

Routine Listings
----------------

TestBase: class
    Base class for tests to start mock & instantiat VS object
parse_args: function
    Parse arguments from mock API call
assert_test: function
    Test pyvesync API calls against existing API
"""
from itertools import zip_longest
import logging
from pathlib import Path
from typing import Any
import yaml
import orjson
from mashumaro.mixins.orjson import DataClassORJSONMixin
from defaults import CALL_API_ARGS, ID_KEYS, API_DEFAULTS

logger = logging.getLogger(__name__)

_SENTINEL = object()


class YAMLWriter:
    """Read and Write API request data to YAML files.

    Arguments
    ---------
    module : str
        name of module that is being tested
    dev_type : str
        device type being tested

    Attributes
    ----------
    self.file_path : Path
        Path to YAML directory, default to API dir in tests folder
    self.file : Path
        Path to YAML file based on device type and module
    self.existings_yaml : dict
        Existing YAML data read to dict object
    self._existing_api : dict, optional
        Existing data dict of a specific API call

    Methods
    -------
    self.existing_api()
        Return existing data dict of a specific API call or None
    self.write_api(api, data, overwrite=False)
        Writes data to YAML file for a specific API call,
        set overwrite=True to overwrite existing data

    """

    def __init__(self, module, dev_type):
        """Init the YAMLWriter class.

        Arguments
        ----------
        module : str
            name of module that is being tested
        dev_type : str
            device type being tested
        """
        self.file_path = self._get_path(module)
        self.file = Path.joinpath(self.file_path, dev_type + '.yaml')
        self.new_file = self._new_file()
        self.existing_yaml = self._get_existing_yaml()
        self._existing_api = None

    @staticmethod
    def _get_path(module) -> Path:
        yaml_dir = Path.joinpath(Path(__file__).parent, 'api', module)
        if not yaml_dir.exists():
            yaml_dir.mkdir(parents=True)
        return yaml_dir

    def _new_file(self) -> bool:
        if not self.file.exists():
            logger.debug(f'Creating new file {self.file}')
            self.file.touch()
            return True
        return False

    def _get_existing_yaml(self) -> Any:
        if self.new_file:
            return None
        with open(self.file, 'rb') as f:
            data = yaml.full_load(f)
        return data

    def existing_api(self, method) -> bool:
        """Check YAML file for existing data for API call.

        Arguments
        ----------
        method : str
            Name of method being tested

        Returns
        -------
        dict or None
            Existing data for API call or None
        """
        if self.existing_yaml is not None:
            current_dict = self.existing_yaml.get(method)
            self._existing_api = current_dict
            if current_dict is not None:
                logger.debug(f'API call {method} already exists in {self.file}')
                return True
        return False

    def write_api(self, method, yaml_dict, overwrite=False):
        """Write API data to YAML file.

        Arguments
        ----------
        method : str
            Name of method being tested
        yaml_dict : dict
            Data to write to YAML file
        overwrite : bool, optional
            Overwrite existing data, default to False
        """
        if self.existing_yaml is not None:
            current_dict = self.existing_yaml.get(method)
            if current_dict is not None and overwrite is False:
                logger.debug(f'API call {method} already exists in {self.file}')
                return
            self.existing_yaml[method] = yaml_dict
        else:
            self.existing_yaml = {method: yaml_dict}
        with open(self.file, 'w', encoding='utf-8') as f:
            yaml.dump(dict(sorted(self.existing_yaml.items())), f, encoding='utf-8')


def api_scrub(api_dict, device_type=None):
    """Recursive function to scrub all sensitive data from API call.

    Arguments
    ----------
    api_dict : dict
        API call data to scrub
    device_type : str, optional
        Device type to use for default values

    Returns
    -------
    dict
        Scrubbed API call data
    """
    def id_cleaner(key, value):
        if key.upper() in ID_KEYS:
            return f"{device_type or 'UNKNOWN'}-{key.upper()}"
        if key in API_DEFAULTS:
            return API_DEFAULTS[key]
        return value

    def nested_dict_iter(nested, mapper, last_key=None):
        if isinstance(nested, dict):
            if nested.get('deviceType') is not None:
                nonlocal device_type
                device_type = nested['deviceType']
            out = {}
            for key, value in nested.items():
                out[key] = nested_dict_iter(value, mapper, key)
            return out

        if isinstance(nested, list):
            return [nested_dict_iter(el, mapper, last_key) for el in nested]
        if not last_key:
            return nested
        return mapper(last_key, nested)
    return nested_dict_iter(api_dict, id_cleaner)


def parse_args(mock_api):
    """Parse arguments from mock API call.

    Arguments
    ----------
    mock_api : mock
        Mock object used to path call_api() method

    Returns
    -------
    dict
        dictionary of all call_api() arguments
    """
    call_args = mock_api.call_args.args
    call_kwargs = mock_api.call_args.kwargs
    all_kwargs = dict(zip(CALL_API_ARGS, call_args))
    all_kwargs.update(call_kwargs)
    if isinstance(all_kwargs.get("json_object"), DataClassORJSONMixin):
        all_kwargs["json_object"] = orjson.loads(all_kwargs["json_object"].to_json())
    elif isinstance(all_kwargs.get("json_object"), dict):
        all_kwargs["json_object"] = orjson.loads(orjson.dumps(all_kwargs["json_object"]))
    return all_kwargs


def assert_test(test_func, all_kwargs, dev_type=None,
                write_api=False, overwrite=False):
    """Test pyvesync API calls against existing API.

    Set `write_api=True` to True to write API call data to YAML file.
    This will not overwrite existing data unless overwrite is True.
    The overwrite argument is only used when API changes, defaults
    to false for development testing. `overwrite=True` and `write_api=True`
    need to both be set to overwrite existing data.

    Arguments
    ----------
    test_func : method
        Method that is being tested
    all_kwargs : dict
        Dictionary of call_api() arguments
    dev_type : str, optional
        Device type being tested
    write_api : bool, optional
        Write API call data to YAML file, default to False
    overwrite : bool, optional
        Overwrite existing data ONLY USE FOR CHANGING API,
          default to False. Must be set with `write_api=True`

    Returns
    -------
    bool
        True if test passes, False if test fails

    Asserts
    -------
        Asserts that the API call data matches the expected data
    """
    if all_kwargs.get('json_object') is not None:
        all_kwargs['json_object'] = api_scrub(all_kwargs['json_object'], dev_type)
    if all_kwargs.get('headers') is not None:
        all_kwargs['headers'] = api_scrub(all_kwargs['headers'], dev_type)
    mod = test_func.__self__.__module__.split(".")[-1]
    if dev_type is None:
        cls_name = test_func.__self__.__class__.__name__
    else:
        cls_name = dev_type
    method_name = test_func.__name__
    writer = YAMLWriter(mod, cls_name)
    if overwrite is True and write_api is True:
        writer.write_api(method_name, all_kwargs, overwrite)
        return True
    if writer.existing_api(method_name) is False:
        logger.warning("No existing, API data for %s %s %s", mod, cls_name, method_name)
        if write_api is True:
            logger.debug("Writing API data for %s %s %s", mod, cls_name, method_name)
            writer.write_api(method_name, all_kwargs, overwrite)
            return True
        else:
            logger.debug("Not writing API data for %s %s %s", mod, cls_name, method_name)
    if writer._existing_api is None:
        return False
    assert writer._existing_api == all_kwargs, dicts_equal(writer._existing_api, all_kwargs)
    return True


def deep_diff(a: Any, b: Any, path: str = ""):
    """Yield dicts describing differences between two (possibly nested) values."""
    if type(a) is not type(b):
        yield {
            "path": path, "change": "type",
            "from": a, "to": b,
            "from_type": type(a).__name__, "to_type": type(b).__name__,
        }
        return

    if isinstance(a, dict):
        keys = set(a) | set(b)
        for k in sorted(keys, key=str):
            p = f"{path}.{k}" if path else str(k)
            if k not in b:
                yield {"path": p, "change": "removed", "value": a[k]}
            elif k not in a:
                yield {"path": p, "change": "added", "value": b[k]}
            else:
                yield from deep_diff(a[k], b[k], p)

    elif isinstance(a, (list, tuple)):
        for i, (va, vb) in enumerate(zip_longest(a, b, fillvalue=_SENTINEL)):
            p = f"{path}[{i}]"
            if va is _SENTINEL:
                yield {"path": p, "change": "added", "value": vb}
            elif vb is _SENTINEL:
                yield {"path": p, "change": "removed", "value": va}
            else:
                yield from deep_diff(va, vb, p)

    elif isinstance(a, set):
        if a != b:
            # repr-sort to avoid TypeError on mixed-type sets
            added = sorted(b - a, key=repr)
            removed = sorted(a - b, key=repr)
            yield {"path": path, "change": "set", "added": added, "removed": removed}

    elif a != b:
        yield {"path": path, "change": "modified", "from": a, "to": b}


def format_diffs(diffs) -> str:
    sym = {"added": "+", "removed": "-", "modified": "~", "type": "±", "set": "∈"}
    out = []
    for d in diffs:
        p = d["path"] or "<root>"
        c = d["change"]
        if c in ("added", "removed"):
            out.append(f"{sym[c]} {p}: {d['value']!r}")
        elif c == "modified":
            out.append(f"{sym[c]} {p}: {d['from']!r} -> {d['to']!r}")
        elif c == "type":
            out.append(f"{sym[c]} {p}: {d['from_type']} -> {d['to_type']} "
                       f"({d['from']!r} -> {d['to']!r})")
        elif c == "set":
            out.append(f"{sym[c]} {p}: +{d['added']!r} -{d['removed']!r}")
    return "\n".join(out)


def dicts_equal(a: dict, b: dict, *, show_diff: bool = True):
    """Return (equal_bool, diffs). Optionally print a readable diff."""
    diffs = list(deep_diff(a, b))
    if show_diff and diffs:
        # print(format_diffs(diffs))
        return 'Differences found:\n' + format_diffs(diffs)
    return "No differences found"


def print_dataclass(d: dict[str, Any]) -> None:
    """Prints dataclass(es) from nested dictionary."""
    def py_type(v: Any) -> str:
        if isinstance(v, bool):
            return "bool"
        if isinstance(v, int):
            return "int"
        if isinstance(v, float):
            return "float"
        if isinstance(v, str):
            return "str"
        if isinstance(v, list):
            return "list[Any]"
        if isinstance(v, dict):
            return "dict[str, Any]"
        return "Any"

    def make_class_name(prefix: str) -> str:
        parts = prefix.split("_")
        return "".join(p.capitalize() for p in parts if p)

    def print_dataclass(name: str, mapping: dict[str, Any]) -> None:
        print("@dataclass")
        print(f"class {name}:")
        if not mapping:
            print("    pass")
            return

        nested: list[tuple[str, dict[str, Any]]] = []

        for key, value in mapping.items():
            if isinstance(value, dict):
                nested_name = make_class_name(f"{name}_{key}")
                print(f"    {key}: {nested_name}")
                nested.append((nested_name, value))
            else:
                print(f"    {key}: {py_type(value)}")

        for nested_name, nested_dict in nested:
            print()
            print_dataclass(nested_name, nested_dict)

    print_dataclass("LehB381s", d)



"""Home and Rooms request and response models.

WORK IN PROGRESS - NOT IMPLEMENTED YET

Dataclasses should follow the naming convention of Request/Response + <API Name> + Model.
Internal models should be named starting with IntResp/IntReq<API Name>Model.

Attributes:
    ResponseLoginResultModel: Model for the 'result' field in login response.
    ResponseLoginModel: Model for the login response.
    RequestLoginModel: Model for the login request.

Notes:
    All models should inherit `ResponseBaseModel` or `RequestBaseModel`. Use
    `pyvesync.models.base_models.DefaultValues` to set default values.

"""

from __future__ import annotations

from dataclasses import dataclass, field

from pyvesync.models.base_models import (
    DefaultValues,
    RequestBaseModel,
    ResponseCodeModel,
)


@dataclass
class RequestHomeModel(RequestBaseModel):
    """Request model for home data.

    Inherits from `RequestVeSyncInstanceMixin` to populate fields from the
    VeSync instance. The `VeSync` instance is passed as a
    keyword argument `RequestHomeModel(manager=instance)`.
    """

    # Arguments set by Manager Instance by passing kw_argument manager
    # these fields should be set to init=False
    accountID: str
    token: str
    userCountryCode: str
    # Non-default constants
    method: str = 'getHomeList'
    # default values
    acceptLanguage: str = DefaultValues.acceptLanguage
    appVersion: str = DefaultValues.appVersion
    timeZone: str = DefaultValues.timeZone
    phoneBrand: str = DefaultValues.phoneBrand
    phoneOS: str = DefaultValues.phoneOS
    traceId: str = field(default_factory=DefaultValues.traceId)
    debugMode: bool = DefaultValues.debugMode


@dataclass
class ResponseHomeModel(ResponseCodeModel):
    """Model for the home data response.

    Inherits from `ResponseCodeModel`. The `ResponseCodeModel` class provides the
    `code` and `msg` fields. The `ResponseHomeModel` class provides the `result`
    field containing the home data.

    Attributes:
        result: dict: The home data.
    """

    result: IntResponseHomeResultModel | IntResponseHomeInfoResultModel | None


@dataclass
class RequestHomeInfoModel(RequestBaseModel):
    """Request model for home room information.

    Inherits from `RequestVeSyncInstanceMixin` to populate fields from the
    VeSync instance. The `VeSync` instance is passed as a
    keyword argument `RequestHomeModel(manager=instance)`.
    """

    # argument to pass in as positional or keyword argument homeId
    homeId: str
    # Arguments set by Manager Instance by passing kw_argument manager
    # these fields should be set to init=False
    accountID: str = field(init=False)
    token: str = field(init=False)
    userCountryCode: str = field(init=False)
    # Non-default constants
    method: str = 'getHomeDetail'
    # default values
    acceptLanguage: str = DefaultValues.acceptLanguage
    appVersion: str = DefaultValues.appVersion
    timeZone: str = DefaultValues.timeZone
    phoneBrand: str = DefaultValues.phoneBrand
    phoneOS: str = DefaultValues.phoneOS
    traceId: str = field(default_factory=DefaultValues.traceId)
    debugMode: bool = DefaultValues.debugMode


@dataclass
class ResponseHomeInfoModel(ResponseCodeModel):
    """Model for the home room information response.

    Inherits from `ResponseCodeModel`. The `ResponseCodeModel` class provides the
    `traceId`, `code`, and `msg` fields. The `ResponseHomeInfoModel` class provides
    the `result` field containing the home room data.

    Attributes:
        result: dict: The home room data.
    """

    result: IntResponseHomeInfoResultModel


@dataclass
class IntResponseHomeInfoResultModel:
    """Internal model for the 'result' field in home room response."""

    roomInfoList: list[IntResponseHomeListModel]


@dataclass
class IntResponseRoomListModel:
    """Internal model for the 'roomList' field in home response."""

    roomID: str
    roomName: str
    deviceList: list[IntResponseRoomDeviceListModel]
    plantComformHumidityRangeLower: int | None = None
    plantComformHumidityRangeHigher: int | None = None
    group_list: list[None] = field(default_factory=list)  # TODO: Add group model


@dataclass
class IntResponseRoomDeviceListModel:
    """Internal model for the device list in room response."""

    logicalDeviceType: int
    virDeviceType: int
    cid: str
    uuid: str
    subDeviceNo: int
    deviceName: str
    configModule: str
    deviceRegion: str
    deviceType: str
    type: str
    connectionType: str
    currentFirmwareVersion: str
    deviceStatus: str
    connectionStatus: str
    mode: str | None = None
    subDeviceType: str | None = None
    macid: str | None = None


@dataclass
class IntResponseHomeResultModel:
    """Internal model for the 'result' field in home response."""

    homeList: list[IntResponseHomeListModel]


@dataclass
class IntResponseHomeListModel:
    """Internal model for the 'homeList' field in home response result."""

    homeId: int
    homeName: str
    nickname: str | None = None
